//--------------------------------------------------------------------------
//
//                     XMM Science Analysis System (SAS)
//                    (c) 1997-2010 European Space Agency
//
//--------------------------------------------------------------------------
//
//	NumericFunctors.cc --- implementation of numeric operator classes
//
//	DESCRIPTION:
//		Defines of functions bound to the operators recognized
//		by the expression scanner. For a list of operators/functions,
//		see Selector.h
//
//	CHANGE HISTORY:
//		UL, 1997-12-17: 1.0.0: - first version
//		UL, 1998-04-09: 1.1.0: - ifthenelse added
//		UL, 1998-05-19: 2.0.0: - major overhaul: look more C++'ish now;
//							     evaluation functions are now static
//		UL, 1998-06-18: 2.1.0: - added classes to support
//							     new 'IN' operator (e.g. PHA in [2:10])
//		UL, 1998-07-28: 2.2.0: - bug fix in FpolygonLogic()
//		UL, 1998-07-29: 2.3.0: - number of polygon vertices no longer limited
//		UL, 1998-10-16: 2.4.0: - bug fix in 'Fpolygon' (local vector
//								 improperly dimensioned - caused crash on
//								 Dec Alpha)
//							   - change implementation of bit complement
//								 operator to make it work on 64-bit
//								 architectures (Dec Alpha)
//		UL, 1999-08-16: 2.5.0: - added hyperbolic trigs +
//								 pow/ceil/floor/fmod/modf/int
//		UL, 2000-02-23: 2.6.0: - in point filter: round to nearest integer
//								 before comparing coordinates
//		UL, 2000-07-03: 2.7.0: - NanOp() added
//
//	AUTHOR:
//		U. Lammers (UL), Dec 17 11:48:12 MET DST 1997
//
//	$Header: /data/xmm02/cvsadmin/soc_repos/sas/packages/selectlib/src/selcore/NumericFunctors.cc,v 1.2 2001/08/10 16:14:56 ulammers Exp $
//

#include <cmath>

#include "basicdefs.h"
#include "Selector.h"
#include "NumericFunctors.h"
#include "SelectError.h"
#include "RegionFilter.h"

using namespace SelectError;

//
//	true if |v1-v2|/|v1| <= tolerance
//
static bool isClose(double v1, double v2, double tolerance)
{
	if (tolerance < 0)
		// tolerance is absolute
		return std::fabs(v1-v2) < -tolerance;
	else {
		// tolerance is relative
		if (v1 == 0.)
			v1 = .1 * tolerance;

		return std::fabs(v1-v2)/fabs(v1) <= tolerance;
	}
} // isClose()


//
//	constructors/destructors for operator classes
//
#define DESTR(op)													\
op::~op() {}

static double EvalLogicAsArith(OpNode me)
{ return me->logic(me) ? 1. : 0.; }

#define OPCLASS(op)													\
DESTR(op)															\
op::op() { registerLogicFunc( op##Logic );							\
		   registerArithFunc(EvalLogicAsArith); }

#define OPCLASS1_L(op)												\
DESTR(op)															\
op::op() { registerLogicFunc( op##Logic );							\
		   registerArithFunc(EvalLogicAsArith); }					\
op::op(const OpNode n1) { OpNode nv[] = { n1 }; addChildren(1, nv);	\
						registerLogicFunc( op##Logic );				\
			 		    registerArithFunc(EvalLogicAsArith); }

#define OPCLASS_INT(op)												\
DESTR(op)															\
op::op() { registerLogicFuncArg( op##LogicArg ); }

#define OPCLASS1_L_INT(op)											\
DESTR(op)															\
op::op() { registerLogicFuncArg( op##LogicArg ); }					\
op::op(const OpNode n1) { OpNode nv[] = { n1 }; addChildren(1, nv);	\
registerLogicFuncArg( op##LogicArg ); }

#define OPCLASS1_A(op)												\
DESTR(op)															\
op::op() { registerArithFunc( op##Arith ); }						\
op::op(const OpNode n1) { OpNode nv[] = { n1 }; addChildren(1, nv);	\
registerArithFunc( op##Arith ); }


#define OPCLASS2_A(op)												\
DESTR(op)															\
op::op() { registerArithFunc( op##Arith ); }						\
op::op(const OpNode n1, const OpNode n2) {							\
	OpNode nv[] = { n1, n2 }; addChildren(2, nv);					\
registerArithFunc( op##Arith ); }

#define OPCLASS2_L(op)												\
DESTR(op)															\
op::op() { registerLogicFunc( op##Logic );							\
		   registerArithFunc(EvalLogicAsArith); }					\
op::op(const OpNode n1, const OpNode n2) {							\
	OpNode nv[] = { n1, n2 }; addChildren(2, nv);					\
	registerLogicFunc( op##Logic );									\
	registerArithFunc(EvalLogicAsArith); }

#define OPCLASS2_L_INT(op)											\
DESTR(op)															\
op::op() { registerLogicFuncArg( op##LogicArg );					\
		   registerArithFunc(EvalLogicAsArith); }					\
op::op(const OpNode n1, const OpNode n2) {							\
	OpNode nv[] = { n1, n2 }; addChildren(2, nv);					\
registerLogicFuncArg( op##LogicArg );								\
registerArithFunc(EvalLogicAsArith); }

#define OPCLASS3(op)												\
DESTR(op)															\
op::op() { registerLogicFunc( op##Logic );							\
		   registerArithFunc(EvalLogicAsArith); }					\
op::op(const OpNode n1, const OpNode n2, const OpNode n3) {			\
	const OpNode nv[] = { n1, n2, n3 }; addChildren(3, nv);			\
registerLogicFunc( op##Logic );										\
registerArithFunc(EvalLogicAsArith); }

#define OPCLASS4(op)												\
DESTR(op)															\
op::op() { registerLogicFunc( op##Logic );							\
		   registerArithFunc(EvalLogicAsArith); }					\
op::op(const OpNode n1, const OpNode n2, const OpNode n3,			\
	   const OpNode n4) {											\
	const OpNode nv[] = { n1, n2, n3, n4 }; addChildren(4, nv);		\
registerLogicFunc( op##Logic );										\
registerArithFunc(EvalLogicAsArith); }

#define OPCLASS5(op)												\
DESTR(op)															\
op::op() { registerLogicFunc( op##Logic );							\
		   registerArithFunc(EvalLogicAsArith); }					\
op::op(const OpNode n1, const OpNode n2, const OpNode n3,			\
	   const OpNode n4, const OpNode n5) {							\
	const OpNode nv[] = { n1, n2, n3, n4, n5 }; addChildren(5, nv);	\
registerLogicFunc( op##Logic );										\
registerArithFunc(EvalLogicAsArith); }

#define OPCLASS6(op)												\
DESTR(op)															\
op::op() { registerLogicFunc( op##Logic );							\
		   registerArithFunc(EvalLogicAsArith); }					\
op::op(const OpNode n1, const OpNode n2, const OpNode n3,			\
	   const OpNode n4,	const OpNode n5, const OpNode n6) {			\
	const OpNode nv[] = { n1, n2, n3, n4, n5, n6 }; addChildren(6, nv);\
registerLogicFunc( op##Logic );										\
registerArithFunc(EvalLogicAsArith); }

#define OPCLASS7(op)												\
DESTR(op)															\
op::op() { registerLogicFunc( op##Logic );							\
		   registerArithFunc(EvalLogicAsArith); }					\
op::op(const OpNode n1, const OpNode n2, const OpNode n3,			\
	   const OpNode n4,	const OpNode n5, const OpNode n6,			\
	   const OpNode n7) {											\
	const OpNode nv[] = { n1, n2, n3, n4, n5, n6, n7 };				\
	addChildren(7, nv);												\
registerLogicFunc( op##Logic );										\
registerArithFunc(EvalLogicAsArith); }

#define OPCLASS10(op)												\
DESTR(op)															\
op::op() { registerLogicFunc( op##Logic );							\
		   registerArithFunc(EvalLogicAsArith); }					\
op::op(const OpNode n1, const OpNode n2, const OpNode n3,			\
	   const OpNode n4, const OpNode n5, const OpNode n6,			\
	   const OpNode n7, const OpNode n8, const OpNode n9,			\
	   const OpNode n10) {											\
	const OpNode nv[] = { n1, n2, n3, n4, n5, n6, n7, n8, n9, n10 };\
	addChildren(10, nv);											\
registerLogicFunc( op##Logic );										\
registerArithFunc(EvalLogicAsArith); }

//
//	evalulation function for binary/unary operators
//	..._EVAL_AA	: arithmetic context for result and children evaluation
//	..._EVAL_LA : arithmetic context for children evaluation / logic
//				  context for result
//	..._EVAL_LL : logic context for resut and children evaluation 

#define BINOP_EVAL_AA(op, operand)									\
double op##Arith(OpNode me)											\
{ return me->child(0)->arith(me->child(0)) operand					\
			me->child(1)->arith(me->child(1)); }

#define BINOP_EVAL_AA_CONV(op, operand, optype)                     \
double op##Arith(OpNode me)											\
{ return DBL(static_cast<optype>(me->child(0)->arith(me->child(0))) \
			operand													\
			static_cast<optype>(me->child(1)->arith(me->child(1)))); }

#define BINOP_EVAL_LA(op, operand)									\
bool op##Logic(OpNode me)											\
{ return me->child(0)->arith(me->child(0)) operand					\
			me->child(1)->arith(me->child(1)); }

#define BINOP_EVAL_LA_ARG(op, operand)								\
bool op##LogicArg(OpNode me, double arg)							\
{ return me->child(0)->arith(me->child(0)) operand arg; }			\

#define BINOP_EVAL_LL(op, operand)									\
bool op##Logic(OpNode me)											\
{ return me->child(0)->logic(me->child(0)) operand					\
			me->child(1)->logic(me->child(1)); }

#define TRIOP_EVAL_LAA_ARG(op, operand1, operand2)					\
bool op##LogicArg(OpNode me, double arg)							\
{ return (me->child(0)->arith(me->child(0)) operand1 arg) &&		\
		 (me->child(1)->arith(me->child(1)) operand2 arg); }

#define UNIOP_EVAL_A(op, operand)									\
double op##Arith(OpNode me)											\
{ return operand me->child(0)->arith(me->child(0)); }

#define UNIOP_EVAL_L(op, operand)									\
bool op##Logic(OpNode me)											\
{ return operand me->child(0)->logic(me->child(0)); }

//
//	evaluator for arithmetic functions
//
#define ARITHFUNC_EVAL(op, funcname, argfac, resfac)				\
double op##Arith(OpNode me)											\
{ return DBL(funcname(me->child(0)->arith(me->child(0)) argfac)) resfac; }

//
//	------------------------( binary operators )--------------------------
//
BINOP_EVAL_AA(Add, +) OPCLASS2_A(Add)
BINOP_EVAL_AA(Sub, -) OPCLASS2_A(Sub)
BINOP_EVAL_AA(Mul, *) OPCLASS2_A(Mul)
static double DivArith(OpNode me)
{
	double denom = me->child(1)->arith(me->child(1));
	if (denom != 0.)
		return me->child(0)->arith(me->child(0)) / denom;
	else
		selerrstr<<"Division by zero in evaluation of selection expression "
				   " - will return 0."<<selwarning("DivisionByZero");

	return 0.;
} // Div::arith()
OPCLASS2_A(Div)

BINOP_EVAL_AA_CONV(Modulo, %, int)
OPCLASS2_A(Modulo)

static double ExpoArith(OpNode me)
{
  	return pow(me->child(0)->arith(me->child(0)),
			   me->child(1)->arith(me->child(1)));
} // Expo::arith()
OPCLASS2_A(Expo)

BINOP_EVAL_AA_CONV(And, &, unsigned long); OPCLASS2_A(And);
BINOP_EVAL_AA_CONV(Or, |, unsigned long); OPCLASS2_A(Or); 
BINOP_EVAL_AA_CONV(Xor, ^, unsigned long); OPCLASS2_A(Xor); 
BINOP_EVAL_AA_CONV(Lshift, <<, unsigned long); OPCLASS2_A(Lshift); 
BINOP_EVAL_AA_CONV(Rshift, >>, unsigned long); OPCLASS2_A(Rshift); 
static double BitcomplArith(OpNode me)
{
	return DBL(~UNT(me->child(0)->arith(me->child(0))));
} // Bitcompl::arith()
OPCLASS1_A(Bitcompl);


bool IntTrueLogicArg(OpNode, double) { return true; } OPCLASS_INT(IntTrue);
bool IntFalseLogicArg(OpNode, double) { return false; } OPCLASS_INT(IntFalse);

BINOP_EVAL_LA_ARG(IntEq, ==); OPCLASS1_L_INT(IntEq);
BINOP_EVAL_LA_ARG(IntLe, <=); OPCLASS1_L_INT(IntLe);
BINOP_EVAL_LA_ARG(IntLt, <); OPCLASS1_L_INT(IntLt);
BINOP_EVAL_LA_ARG(IntGe, >=); OPCLASS1_L_INT(IntGe);
BINOP_EVAL_LA_ARG(IntGt, >); OPCLASS1_L_INT(IntGt);

TRIOP_EVAL_LAA_ARG(IntCC, <=, >=); OPCLASS2_L_INT(IntCC);
TRIOP_EVAL_LAA_ARG(IntOC, <, >=); OPCLASS2_L_INT(IntOC);
TRIOP_EVAL_LAA_ARG(IntCO, <=, >); OPCLASS2_L_INT(IntCO);
TRIOP_EVAL_LAA_ARG(IntOO, <, >); OPCLASS2_L_INT(IntOO);

BINOP_EVAL_LL(AndIf, &&); OPCLASS2_L(AndIf);
BINOP_EVAL_LL(OrIf, ||); OPCLASS2_L(OrIf); 

BINOP_EVAL_LA(Lt, <); OPCLASS2_L(Lt); 
BINOP_EVAL_LA(Le, <=); OPCLASS2_L(Le);
BINOP_EVAL_LA(Gt, >); OPCLASS2_L(Gt);
BINOP_EVAL_LA(Ge, >=); OPCLASS2_L(Ge); 
BINOP_EVAL_LA(Eq, ==); OPCLASS2_L(Eq); 
BINOP_EVAL_LA(Ne, !=); OPCLASS2_L(Ne); 

UNIOP_EVAL_L(Not, !); OPCLASS1_L(Not);
UNIOP_EVAL_A(Uminus, -); OPCLASS1_A(Uminus);

//
//	-------------------------( arithmetic functions )-----------------------
//
ARITHFUNC_EVAL(Fsin, sin, *DEG_TO_RAD,); OPCLASS1_A(Fsin);
ARITHFUNC_EVAL(Fcos, cos, *DEG_TO_RAD,); OPCLASS1_A(Fcos);
ARITHFUNC_EVAL(Ftan, tan, *DEG_TO_RAD,); OPCLASS1_A(Ftan);
ARITHFUNC_EVAL(Arcsin, asin, , *RAD_TO_DEG); OPCLASS1_A(Arcsin);
ARITHFUNC_EVAL(Arccos, acos, , *RAD_TO_DEG); OPCLASS1_A(Arccos);
ARITHFUNC_EVAL(Arctan, atan, , *RAD_TO_DEG); OPCLASS1_A(Arctan);
static double Arctan2Arith(OpNode me)
{
	return atan2(me->child(0)->arith(me->child(0)),
				 me->child(1)->arith(me->child(1)))	* RAD_TO_DEG;
} // Arctan2::arith()
OPCLASS2_A(Arctan2);
ARITHFUNC_EVAL(Fsinh, sinh, ,); OPCLASS1_A(Fsinh);
ARITHFUNC_EVAL(Fcosh, cosh, ,); OPCLASS1_A(Fcosh);
ARITHFUNC_EVAL(Ftanh, tanh, ,); OPCLASS1_A(Ftanh);

ARITHFUNC_EVAL(Fabs, fabs, ,); OPCLASS1_A(Fabs);
ARITHFUNC_EVAL(Fexp, exp, ,); OPCLASS1_A(Fexp);
ARITHFUNC_EVAL(Flog, log, ,); OPCLASS1_A(Flog);
ARITHFUNC_EVAL(Flog10, log10, ,); OPCLASS1_A(Flog10); 
ARITHFUNC_EVAL(Fsqrt, sqrt, ,); OPCLASS1_A(Fsqrt);
static double FintArith(OpNode me)
{
	return INT(me->child(0)->arith(me->child(0)));
} // IntArith::arith()
OPCLASS1_A(Fint);
ARITHFUNC_EVAL(Ceil, ceil, ,); OPCLASS1_A(Ceil);
ARITHFUNC_EVAL(Floor, floor, ,); OPCLASS1_A(Floor);
static double PowArith(OpNode me)
{
	return pow(me->child(0)->arith(me->child(0)),
			   me->child(1)->arith(me->child(1)));
} // Pow::Arith()
OPCLASS2_A(Pow);
static double ModfArith(OpNode me)
{
	double dum;
	return modf(me->child(0)->arith(me->child(0)), &dum);
} // Modf::Arith()
OPCLASS1_A(Modf);
static double FmodArith(OpNode me)
{
	return fmod(me->child(0)->arith(me->child(0)),
			    me->child(1)->arith(me->child(1)));
} // Fmod::Arith()
OPCLASS2_A(Fmod);

//
//	--------------------( spatial selection functions )---------------------
//
inline static int nearestInt(double x)
{
	return INT(x >= 0 ? std::floor(x + .5) : std::ceil(x - 0.5));
}

static bool FpointLogic(OpNode me)
{
	double x	= me->child(0)->arith(me->child(0)),	// x coordinate
		   y	= me->child(1)->arith(me->child(1)),	// y coordinate
		   xp	= me->child(2)->arith(me->child(2)),	// RAWX
		   yp	= me->child(3)->arith(me->child(3));	// RAWY

	return nearestInt(x) == nearestInt(xp) && nearestInt(y) == nearestInt(yp);
} // Fpoint::logic()
OPCLASS4(Fpoint);


static bool FlineLogic(OpNode me)
{
	double	x1	= me->child(0)->arith(me->child(0)),	// start point x
			y1	= me->child(1)->arith(me->child(1)),	// start point y
			x2	= me->child(2)->arith(me->child(2)),	// end point x
			y2	= me->child(3)->arith(me->child(3)),	// end point y
			xp	= me->child(4)->arith(me->child(4)),	// RAWX
			yp	= me->child(5)->arith(me->child(5));	// RAWY

	if (x1 == x2)
		// vertical line
		return xp == x1 && y1<= yp && y2 >= yp;

	double m = (y1 - y2)/(x1 - x2),
		   b = y1 - m * x1;

	return isClose(yp, m * xp + b, -.5); //	absolute deviation of .5 allowed
} // Fline::logic()
OPCLASS6(Fline);


static bool FcircleLogic(OpNode me)
{
	double x		= me->child(0)->arith(me->child(0)),	// center x
		   y		= me->child(1)->arith(me->child(1)),	// center y
		   radius	= me->child(2)->arith(me->child(2)),	// radius
		   rawx		= me->child(3)->arith(me->child(3)),	// RAWX
		   rawy		= me->child(4)->arith(me->child(4));	// RAWY

	rawx -= x;
	rawy -= y;

	return rawx*rawx + rawy*rawy <= radius*radius;
} // Fcircle::logic()
OPCLASS5(Fcircle);

static bool FringLogic(OpNode me)
{
	double x		= me->child(0)->arith(me->child(0)),	// center x
		   y		= me->child(1)->arith(me->child(1)),	// center y
		   radius1	= me->child(2)->arith(me->child(2)),	// r inner circle
		   radius2	= me->child(3)->arith(me->child(3)),	// r outer circle
		   rawx		= me->child(4)->arith(me->child(4)),	// RAWX
		   rawy		= me->child(5)->arith(me->child(5));	// RAWY

	rawx -= x;
	rawy -= y;
	double dist2 = rawx*rawx + rawy*rawy;

	return dist2 >= radius1*radius1 && dist2 <= radius2*radius2;
} // Fring::logic()
OPCLASS6(Fring);


static bool insideEllipse(double x, double y, double x_half, double y_half,
						  double phi, double rawx, double rawy)
{
	rawx -= x;
	rawy -= y;
	if (phi != 0.) {
		//
		//	transform point into frame aligned with the box - (0/0) is
		//	center of ellipse
		//
		double sin_phi = std::sin(phi * DEG_TO_RAD),
			   cos_phi = std::cos(phi * DEG_TO_RAD), old_rawx;

		old_rawx = rawx;
		rawx = old_rawx * cos_phi + rawy * sin_phi;
		rawy = -old_rawx * sin_phi + rawy * cos_phi;
	}

	return rawx*rawx/(x_half*x_half) + rawy*rawy/(y_half*y_half) <= 1;
} // insideEllipse


static bool FellipseLogic(OpNode me)
{
	return insideEllipse(me->child(0)->arith(me->child(0)),		// center x
						 me->child(1)->arith(me->child(1)),		// center y
						 me->child(2)->arith(me->child(2)),		// half width x
						 me->child(3)->arith(me->child(3)),		// half width y
						 me->child(4)->arith(me->child(4)),		// rot. angle
						 me->child(5)->arith(me->child(5)),		// RAWX
						 me->child(6)->arith(me->child(6)));	// RAWY
} // Fellipse::logic()
OPCLASS7(Fellipse);


static bool FrellipseLogic(OpNode me)
{
	double x = me->child(0)->arith(me->child(0)),			// center x
		   y = me->child(1)->arith(me->child(1)),			// center y
		   xhalfInner = me->child(2)->arith(me->child(2)),	// major axis
															// inner ellipse
		   yhalfInner = me->child(3)->arith(me->child(3)),	// minor axis
															// inner ellipse
		   xhalfOuter = me->child(4)->arith(me->child(4)),	// major axis
															// outer ellipse
		   yhalfOuter = me->child(5)->arith(me->child(5)),	// minor axis
															// outer ellipse
		   rotAngInner = me->child(6)->arith(me->child(6)),	// rot. angle
															// inner ellipse
		   rotAngOuter = me->child(7)->arith(me->child(7)),	// rot. angle
															// outer ellipse
		   rawX	= me->child(8)->arith(me->child(8)),		// RAWX
		   rawY = me->child(9)->arith(me->child(9));		// RAWY

	return insideEllipse(x, y, xhalfOuter, yhalfOuter, rotAngOuter,
						 rawX, rawY) &&
			!insideEllipse(x, y, xhalfInner, yhalfInner, rotAngInner,
						 rawX, rawY);
} // Frellipse::logic()
OPCLASS10(Frellipse);


static bool FboxLogic(OpNode me)
{
	double x		= me->child(0)->arith(me->child(0)),	// center x
		   y		= me->child(1)->arith(me->child(1)),	// center y
		   x_half	= me->child(2)->arith(me->child(2)),	// half width x
		   y_half	= me->child(3)->arith(me->child(3)),	// half width y
		   phi		= me->child(4)->arith(me->child(4)),	// rotation angle
		   rawx		= me->child(5)->arith(me->child(5)),	// RAWX
		   rawy		= me->child(6)->arith(me->child(6));	// RAWY

	rawx -= x;
	rawy -= y;
	if (phi != 0.) {
		//
		//	transform point into frame aligned with the box - (0/0) is
		//	center of box
		//
		double sin_phi = sin(phi * DEG_TO_RAD),
			   cos_phi = cos(phi * DEG_TO_RAD), old_rawx;

		old_rawx = rawx;
		rawx = old_rawx * cos_phi + rawy * sin_phi;
		rawy = -old_rawx * sin_phi + rawy * cos_phi;
	}

	return std::fabs(rawx) <= x_half && std::fabs(rawy) <= y_half;
} // Fbox::logic()
OPCLASS7(Fbox);


//
//	normalize angle in range [-Pi, Pi]
//
static double normAng(double ang)
{
	double norm = std::fmod(ang, 360.);

	if (norm > 180.)
		return norm - 360.;
	else if (norm < -180.)
		return norm + 360.;

	return norm;
} // normAng

static bool FsectorLogic(OpNode me)
{
	double x = me->child(0)->arith(me->child(0)),				// center x
		   y = me->child(1)->arith(me->child(1)),				// center y
		   phi1 = normAng(me->child(2)->arith(me->child(2))),	// normalized
																// first angle
		   phi2 = normAng(me->child(3)->arith(me->child(3))),	// normalized
																// second angle
		   rawx = me->child(4)->arith(me->child(4)),			// test point x
		   rawy = me->child(5)->arith(me->child(5));			// test point y

	rawx -= x;
	rawy -= y;
	if (rawx == 0. && rawy == 0.)
		return true;	// center belong to sector

	double phi = std::atan2(rawy, rawx) * RAD_TO_DEG;

	if (phi1 <= phi2)
		return phi1 <= phi && phi2 >= phi;

	//	sector crosses boundary between quadrants #2 and #3
	//	where angle changes from +Pi -> -Pi
	if (phi > 0.)
		//	phi is in quadrant #1 or #2
		return phi >= phi1;
	else
		//	phi is in quadrant #3 or #4
		return phi <= phi2;
} // Fsector::logic()
OPCLASS6(Fsector);


static bool FrectangleLogic(OpNode me)
{
	double x0	= me->child(0)->arith(me->child(0)),	// lower left corner x
		   y0	= me->child(1)->arith(me->child(1)),	// lower left corner y
		   x1	= me->child(2)->arith(me->child(2)),	// upper right corner x
		   y1	= me->child(3)->arith(me->child(3)),	// upper right corner y
		   phi	= me->child(4)->arith(me->child(4)),	// rotation angle
		   rawx	= me->child(5)->arith(me->child(5)),	// RAWX
		   rawy	= me->child(6)->arith(me->child(6));	// RAWY

	rawx -= x0;
	rawy -= y0;
	if (phi != 0.) {
		//
		//	transform point into frame aligned with the box - (0/0) is
		//	lower left corner of box
		//
		double sin_phi = std::sin(phi * DEG_TO_RAD),
			   cos_phi = std::cos(phi * DEG_TO_RAD);

		double old_rawx = rawx;
		rawx = old_rawx * cos_phi + rawy * sin_phi;
		rawy = -old_rawx * sin_phi + rawy * cos_phi;
	}
	x1 -= x0;
	y1 -= y0;

	return rawx >= 0. && rawx <= x1 && rawy >= 0. && rawy <= y1;
} // Frectangle::logic()
OPCLASS7(Frectangle);


static bool FdiamondLogic(OpNode me)
{
	double x		= me->child(0)->arith(me->child(0)),	// center x
		   y		= me->child(1)->arith(me->child(1)),	// center y
		   xHalf	= me->child(2)->arith(me->child(2)),	// half width x
		   yHalf	= me->child(3)->arith(me->child(3)),	// half width y
		   phi		= me->child(4)->arith(me->child(4)),	// rotation angle
		   rawx		= me->child(5)->arith(me->child(5)),	// RAWX
		   rawy		= me->child(6)->arith(me->child(6));	// RAWY

	rawx -= x;
	rawy -= y;
	if (phi != 0.) {
		//
		//	transform point into frame aligned with the box - (0/0) is
		//	center of box
		//
		double sin_phi = std::sin(phi * DEG_TO_RAD),
			   cos_phi = std::cos(phi * DEG_TO_RAD), old_rawx;

		old_rawx = rawx;
		rawx = old_rawx * cos_phi + rawy * sin_phi;
		rawy = -old_rawx * sin_phi + rawy * cos_phi;
	}

	if (xHalf == 0.)
		return true;

	return std::fabs(rawy) <= -yHalf/xHalf * std::fabs(rawx) + yHalf;
} // Fdiamond::logic()
OPCLASS7(Fdiamond);


static bool FnearLogic(OpNode me)
{
	return isClose(me->child(0)->arith(me->child(0)),
				   me->child(1)->arith(me->child(1)),
				   me->child(2)->arith(me->child(2)));
} // Fnear::logic()
OPCLASS3(Fnear);


static bool IfthenelseLogic(OpNode me)
{
	if (me->child(0)->logic(me->child(0)))
		return me->child(1)->logic(me->child(1));
	else
		return me->child(2)->logic(me->child(2));
} // Ifthenelse::logic()
OPCLASS3(Ifthenelse);


//
//	-----------------------( true/false nodes )------------------------------
//
static bool TrueNodeLogic(OpNode) { return true; }
OPCLASS(TrueNode);

static bool FalseNodeLogic(OpNode) { return false; }
OPCLASS(FalseNode);

//
//	--------------------------( NanOp node )---------------------------------
//
static bool NanOpLogic(OpNode me)
{
	return isnan(me->child(0)->value());
} // NanOpLogic()

OPCLASS1_L(NanOp);

//
//	----------------------( InInterval node )--------------------------------
//
static bool InIntervalLogic(OpNode me)
{
	unsigned nNodes = me->nChildren();

	//	the last node is the argument, evaluate that first
	double arg = me->child(nNodes - 1)->arith(me->child(nNodes - 1));

	for (unsigned n=0; n<nNodes - 1; ++n)
		if (me->child(n)->logicArg(me->child(n), arg))
			return true;

	return false;
} // InIntervalLogic()
OPCLASS(InInterval);
