//--------------------------------------------------------------------------
//
//                     XMM Science Analysis System (SAS)
//                    (c) 1997-2010 European Space Agency
//
//--------------------------------------------------------------------------
//
//	RegionFilter.cc --- implementation of region filter related classes
//
//	DESCRIPTION:
//		Implementation of region filters.
//
//	CHANGE HISTORY:
//		UL, 2001-08-07: 1.0.0: - first version
//
//	AUTHOR:
//		U. Lammers (UL), Tue Aug  7 11:27:22 MET DST 2001
//
//	$Header: /data/xmm02/cvsadmin/soc_repos/sas/packages/selectlib/src/selcore/RegionFilter.cc,v 1.3 2001/08/15 16:44:07 ulammers Exp $
//
#include "RegionFilter.h"
#include "StandardPolygon.h"

#define DESTR(op)													\
op::~op() {}

#define OPCLASS3(op)												\
DESTR(op)															\
op::op() { registerLogicFunc( op##Logic );							\
           registerArithFunc(EvalLogicAsArith); }					\
op::op(const OpNode n1, const OpNode n2, const OpNode n3) {			\
    const OpNode nv[] = { n1, n2, n3 }; addChildren(3, nv);			\
registerLogicFunc( op##Logic );										\
registerArithFunc(EvalLogicAsArith); }


namespace {

//	evaluate logical function in arithmetic context
double EvalLogicAsArith(OpNode me)
{ return me->logic(me) ? 1. : 0.; }

}

//
//	classes for point list
//
void NodePointList::addPoint(const OpNode x, const OpNode y)
{
	this->push_back(NodePoint(x, y));
	setConstness(isConst() && x->isConst() && y->isConst());
} // NodePointList::addPoint()

//
//	------------------------( polygon )-------------------------------------
//
static bool evaluateTriangulatablePolygonFilter(OpNode me)
{
	// evaluate test point
	Point<double> testPoint(me->child(0)->arith(me->child(0)),
							me->child(1)->arith(me->child(1)));

	//	this node is also a TriangulatablePolygon - cast and evaluate
	return dynamic_cast<TriangulatablePolygon *>(me)->contains(testPoint);
} // evaluateTriangulatablePolygonFilter();


TriangulatablePolygonFilter::TriangulatablePolygonFilter(
											const std::vector<double> &xCoords,
											const std::vector<double> &yCoords,
											const OpNode testX,
											const OpNode testY,
											bool testSegments)
		: TriangulatablePolygon(xCoords, yCoords, false, testSegments)
{
	OpNode nd[] = { testX, testY };
	addChildren(2, nd);	//	register test point coordinates as children
	registerLogicFunc(evaluateTriangulatablePolygonFilter);
	registerArithFunc(EvalLogicAsArith);
} // TriangulatablePolygonFilter::TriangulatablePolygonFilter()


TriangulatablePolygonFilter::~TriangulatablePolygonFilter() {}


static bool VariablePolygonFilterLogic(OpNode me)
{
	// evaluate test point
	GeoPoint testPoint = { me->child(1)->arith(me->child(1)),
						   me->child(2)->arith(me->child(2)) };

	//
	//	evalulate corners of polygon - first child of tree
	//	points to list of corner points
	//
	NodePointList *pl = reinterpret_cast<NodePointList *>(me->child(0));	
	unsigned nPoints = pl->size();

	GeoPoint poly1[nPoints+2]; //, poly2[nPoints+2];
			 // +2: Point-in-Polygon routine 'insidePolygon' needs
			 //		these extra elements as work space ;(
	double			minY = 1.e99, minX = 1.e99;
	unsigned		minIndx = 0;
	for (unsigned n=0; n<nPoints; ++n) {
		double x = (*pl)[n].xNode()->arith((*pl)[n].xNode()),
			   y = (*pl)[n].yNode()->arith((*pl)[n].yNode());
		poly1[n+1].x = x;
		poly1[n+1].y = y;
		if (y < minY) {
			minY = y;
			minX = x;
			minIndx = n;
		} else if (y == minY && x < minX) {
			minX = x;
			minIndx = n;
		}
	}

#if 0
	//
	//	construct new polygon such that first point is the one with the
	//	lowest x-coordinates among all points with the lowest y-coordinate
	//
	for (unsigned i=1; i<=nPoints-minIndx+1; ++i)
		poly2[i] = poly1[minIndx+i-1];
	for (unsigned i=1; i<=minIndx-1; ++i)
		poly2[nPoints-minIndx+1+i] = poly1[i];
#endif

	//
	//	polygon is now defined - call point-in-polgon test
	//
	return StandardPolygon::insidePolygon(testPoint, poly1, nPoints);
} // VariablePolygonFilterLogic();

OPCLASS3(VariablePolygonFilter)


namespace PolygonScope {

OpNode newPolyFilter(const OpNode opl,		//	point list
					 const OpNode testX,	//	text point x
					 const OpNode testY,	//	test point y
					 bool testSegments)
{
	if (opl->isConst()) {
		//	point list is const - evaluate and instantiate Triangulatable
		//	polygon
		NodePointList *pl = reinterpret_cast<NodePointList *>(opl);
		unsigned N = pl->size();
		std::vector<double> xCoords(N), yCoords(N);
		for (unsigned n=0; n<N; ++n) {
			xCoords[n] = (*pl)[n].xNode()->arith((*pl)[n].xNode()),
			yCoords[n] = (*pl)[n].yNode()->arith((*pl)[n].yNode());
		}

		TriangulatablePolygonFilter *polyFilt;
		try {
			polyFilt =
				new TriangulatablePolygonFilter(xCoords, yCoords,
												testX, testY, testSegments);
			//	point list no longer needed - delete it
			delete opl;
		} catch (TriangulatablePolygonFilter::IntersectingSegments) {
			//	the polygon has intersting segments - instantiate
			//	conventional polygon filter
			return new VariablePolygonFilter(opl, testX, testY);
		}
		return polyFilt;
	}

	//	we have a variable polygon
	return new VariablePolygonFilter(opl, testX, testY);
} // newPolyFilter()

}

#undef DESTR
#undef OPCLASS3
