//---------------------------------------------------------------------------
//
//                     XMM Science Analysis System (SAS)
//                    (c) 1997-2010 European Space Agency
//
//---------------------------------------------------------------------------
//
//	SelectBlockFuncs.h: --- operator functions which require reading a file
//
//	DESCRIPTION:
//		Below are prototypes for functions called from the bison
//		generated parser code for input expressions:
//			gti(<gtifilename>, timecol)
//			mask(<maskfilename>, xoffset, yoffset, xCol, yCol)
//
//	CHANGE HISTORY:
//		UL, 1998-01-05: 1.0.0: - first version
//		UL, 1998-10-07: 1.1.0: - some modifications to support WCS in the
//								 mask filter
//		UL, 1999-06-28: 1.2.0: - added another version of
//									 RegionFilter::expandRegion() which does
//								 not take X/Y column names as arguments but
//								 gets them from specified region file
//
//	AUTHOR:
//		U. Lammers (UL), Wed Aug 27 11:48:12 MET DST 1997
//
//	$Header: /data/xmm02/cvsadmin/soc_repos/sas/packages/selectlib/src/selcore/SelectBlockFuncs.h,v 1.2 2001/08/10 16:14:56 ulammers Exp $
///

#ifndef SELECT_BLOCK_FUNC_H
#define SELECT_BLOCK_FUNC_H

#include <string>
#include "Selector.h"
#include "NumericFunctors.h"
#include "../utils/SelectWCS.h"

class GtiFilter {
public:
	static const OpNode expandGtis(const std::string &fileName,
								   const OpNode timeNode);
		//
		//	PURPOSE: turn a list of Good Time Intervals in a file into
		//			 a tree representation
		//	REQUIRE: o file named <fileName> must exist
		//			 o <timeNode> must be a valid operator node
		//	PROMISE: RETURN is operator node that can be evaluated in
		//			 logical context; result is true if result of
		//			 arithmetic evaluation of <timeNode> lies within at
		//			 least one of the GTIs
};


class RegionFilter : public OperatorNode {
public:
	static const OpNode expandRegion(const std::string &fileName,
								     const OpNode xNode, const OpNode yNode);
		//
		//	PURPOSE: turn the specification of a selection region contained
		//			 in an ASC region file into a tree representation
		//	REQUIRE: o file named <fileName> must exist and contain
		//			   valid entries (ASC region definitions)
		//			 o <xNode>/<yNode> must be valid operator nodes that can
		//			   be evaluated in arithmeric context
		//	PROMISE: RETURN is operator node that can be evaluated in
		//			 logical context; result is true if point resulting
		//			 from arithmetic evaluation of <xNode>/<yNode> lies
		//			 within the region specified in region file <fileName>

	static const OpNode expandRegion(const std::string &fileName);
		//
		//	PURPOSE: as above
		//	REQUIRE: as above; column names are to be taken from MFORM/MTYPE
		//			 keywords in specified region file
		//	PROMISE: as above
};

class Table;

class MaskFilter : public OperatorNode {
public:
	MaskFilter(const std::string &fileName, const OpNode xoffset,
			   const OpNode yoffset, const OpNode xNode, const OpNode yNode);
		//
		//	PURPOSE: construct filter which can be evaluated in a logical
		//			 context
		//	REQUIRE: o file <fileName> must exist and contain a mask image
		//			 o all nodes passed as arguments must be evaluable
		//			   in arithmetic context
		//	PROMISE: o establish logic pointer in base class to evaluate
		//			   node in logic context; this returns 
		//			   true if point specified via <xNode>/<yNode>
		//			   lies within the mask image offset by <xoffset>/<yoffset>

	unsigned	*arrayData() const { return _arrayData; }
	unsigned	auxData(unsigned n) const { return _extent[n]; }
	const SelectWCS	*maskWCS(unsigned n) const { return &_maskWCS[n]; }
	const SelectWCS	*eventTableWCS(unsigned n) const {
												return &_eventTableWCS[n]; }

	~MaskFilter();

private:
	unsigned	*_arrayData;	//	the mask image data
	unsigned	_extent[2];		//	extent of mask image in X/Y
	SelectWCS	_maskWCS[2],	//	WCS for the mask image ...
				_eventTableWCS[2];
								//	... and for the table;
	std::string		_xcol, _ycol;	//	name of coordinate columns

	void getWCSfromEventTable(const Table *);
};

#endif
