//--------------------------------------------------------------------------
//
//                     XMM Science Analysis System (SAS)
//                    (c) 1997-2010 European Space Agency
//
//--------------------------------------------------------------------------
//
//	StringFunctors.cc --- implementation of string operator classes
//
//	DESCRIPTION:
//		Defines of functions bound to the operators recognized
//		by the expression scanner. For a list of operators/functions,
//		see Selector.h
//
//	CHANGE HISTORY:
//		UL, 2000-06-28: 1.0.0: - first version
//
//	AUTHOR:
//		U. Lammers (UL), Wed Jun 28 16:47:58 MET DST 2000
//
//	$Header: /data/xmm02/cvsadmin/soc_repos/sas/packages/selectlib/src/selcore/StringFunctors.cc,v 1.3 2001/08/15 16:44:08 ulammers Exp $
//

#include <cctype>
#include <cstring>

#include "basicdefs.h"
#include "Selector.h"
#include "NumericFunctors.h"
#include "StringFunctors.h"
#include "SelectError.h"

using namespace SelectError;

//
//	constructors/destructors for operator classes
//
#define DESTR(op)													\
op::~op() {}

static double EvalLogicAsArith(OpNode me)
{ return me->logic(me) ? 1. : 0.; }

#define OPCLASS1_S(op)												\
DESTR(op)															\
op::op() { registerTextFunc( op##Text ); }							\
op::op(const OpNode n1) { OpNode nv[] = { n1 }; addChildren(1, nv);	\
						registerTextFunc( op##Text ); }

#define OPCLASS2_S(op)												\
DESTR(op)															\
op::op() { registerTextFunc( op##Text ); }							\
op::op(const OpNode n1, const OpNode n2) {							\
	OpNode nv[] = { n1, n2 }; addChildren(2, nv);					\
	registerTextFunc( op##Text ); }

#define OPCLASS1_A(op)												\
DESTR(op)															\
op::op() { registerArithFunc( op##Arith ); }						\
op::op(const OpNode n1) { OpNode nv[] = { n1 }; addChildren(1, nv);	\
						registerArithFunc( op##Arith ); }

#define OPCLASS2_L(op)												\
DESTR(op)															\
op::op() { registerArithFunc( EvalLogicAsArith );					\
		   registerLogicFunc( op##Logic ); }						\
op::op(const OpNode n1, const OpNode n2) {							\
	OpNode nv[] = { n1, n2 }; addChildren(2, nv);					\
registerArithFunc( EvalLogicAsArith );								\
registerLogicFunc( op##Logic ); }

#define STRCMP(op, arithOp)											\
static bool op##Logic(OpNode me) {									\
	return strcmp(me->child(0)->textual(me->child(0)),				\
				  me->child(1)->textual(me->child(1))) arithOp 0;	\
}

//
//	Relational/equality operators operators
//
STRCMP(StrEq, ==); OPCLASS2_L(StrEq);
STRCMP(StrNe, !=); OPCLASS2_L(StrNe);
STRCMP(StrLt, <); OPCLASS2_L(StrLt);
STRCMP(StrLe, <=); OPCLASS2_L(StrLe);
STRCMP(StrGt, >); OPCLASS2_L(StrGt);
STRCMP(StrGe, >=); OPCLASS2_L(StrGe);

//
//	Upper/Lower
//
static char *UpperText(OpNode me)
{
	const char *src = me->child(0)->textual(me->child(0));
	char *dest = me->text();
	while (*src)
		*dest++ = toupper(*src++);
	*dest = '\0';

	return me->text();
} // UpperText()

OPCLASS1_S(Upper);

static char *LowerText(OpNode me)
{
	const char *src = me->child(0)->textual(me->child(0));
	char *dest = me->text();
	while (*src)
		*dest++ = tolower(*src++);
	*dest = '\0';

	return me->text();
} // LowerText()

OPCLASS1_S(Lower);

static double StrlenArith(OpNode me)
{
	return std::strlen(me->child(0)->textual(me->child(0)));
} // StrlenArith()

OPCLASS1_A(Strlen);


static double AsciiArith(OpNode me)
{
	return *me->child(0)->textual(me->child(0));
} // AsciiArith()

OPCLASS1_A(Ascii);

//
//	string range operator
//
static char *StrRangeText(OpNode me)
{
	int lo = INT(me->child(0)->arith(me->child(0))),
		hi = INT(me->child(1)->arith(me->child(1)));

	if (lo < 0)
		lo = 0;	//	ignore negative indices
	if (UNT(hi) >= me->textLength())
		hi = me->textLength()-1;

	const char *src = me->child(2)->textual(me->child(2));
	char *dest = me->text();
	int n = MIN(hi-lo+1, INT(me->textLength()));
	std::strncpy(dest, src+lo, n);
	dest[n] = '\0';

	return dest;
} // StrRangeText()


StrRange::StrRange(bool lowerGiven, bool upperGiven, OpNode c1, OpNode c2)
{
	OpNode lower, upper;
	if (!lowerGiven)
		lower = new NumericLiteral(0);	//	[:hi] == [0:hi]
	else
		lower = c1;

	if (!upperGiven)
		upper = new NumericLiteral(1024);	//	[lo:] == [lo:Infinity]
	else
		if (lowerGiven)
			upper = c2;
		else
			upper = c1;

	OpNode nv[] = { lower, upper };
	addChildren(NUM_OF(nv), nv);
	registerTextFunc(StrRangeText);
}

StrRange::~StrRange() {}

//
//	string concatenation
//
static char *StrConcatText(OpNode me)
{
	const char *s1 = me->child(0)->textual(me->child(0)),
			   *s2 = me->child(1)->textual(me->child(1));
	char *dest = me->text();

	std::strcpy(dest, s1);
	return std::strcat(dest, s2);
} // StrConcatText()


OPCLASS2_S(StrConcat);

#undef STRCMP
#undef OPCLASS1_S
#undef OPCLASS2_S
#undef OPCLASS1_A
#undef OPCLASS2_L

#undef DESTR
