//----------------------------------------------------------------------------
//
//                     XMM Science Analysis System (SAS)
//                    (c) 1997-2010 European Space Agency
//
//---------------------------------------------------------------------------
//
//	ASCregion.cc: --- implementation of ASCregion classes
//
//	DESCRIPTION:
//		Instances of class ASCregion represent region filters as
//		specified in "ASC FITS File Designers' Guide", J. McDowell and
//		A. Rots.
//
//	CHANGE HISTORY:
//		UL, 1998-03-25: 1.0.0: - first version
//		UL, 1998-04-06: 1.1.0: - brought in line with 1998-03-10 version of [1]
//							   - change in DAL API affecting way string column
//								 are read
//		UL, 1998-05-19: 1.2.0: - make use of <selerrstr>
//		UL, 1999-03-03: 1.3.0: - modifications enforced by Lowmem Dal
//		UL, 1999-11-03: 1.4.0: - added support for new unsigned types in
//								 dal>=1.80
//		UL, 1999-12-08: 1.5.0: - bug fix in InASCregion: did not treat
//								 columns "R"/"ROTANG" as optional
//		UL, 2000-06-22: 1.6.0: - add terminal vertex when writing out
//								 polygons
//		UL, 2001-07-26: 1.7.0: - case of SHAPE value in CXC input table is
//								 now indifferent (used to require SHAPE
//								 names in uppercase)
//
//	REFERENCES:
//		[1] : FITS REGION Binary Table Design, J. McDowell and A. Rots,
//			  ASC-FITS-REGION-1.0
//
//	AUTHOR:
//		U. Lammers (UL), Wed Mar 25 16:52:23 MET 1998
//
//	$Header: /data/xmm02/cvsadmin/soc_repos/sas/packages/selectlib/src/utils/ASCregion.cc,v 1.2 2001/08/10 16:15:00 ulammers Exp $
//

#include <cstring>
#include "ASCregion.h"
#include "SetTableColumn.h"
#include "SelectError.h"
#include "Dal.h"
#include "common.h"
#include "stringutil.h"

using namespace SelectError;

inline unsigned intColData(const Column *col, unsigned n)
{
	ColumnData *d = col->data();
	unsigned value = 0;
	switch (col->dataType()) {
		case Dal::Int8:  value = d->int8Data()[n]; break;
		case Dal::Int16: value = d->int16Data()[n]; break;
		case Dal::Uint16: value = d->uint16Data()[n]; break;
		case Dal::Int32: value = d->int32Data()[n]; break;
		case Dal::Uint32: value = d->uint32Data()[n]; break;
		default:		 selerrstr<<"Cannot convert Dal data type "
								  <<UNT(col->dataType())<<" in column "
								  <<col->qualifiedName()
								  <<"to unsigned"
								  <<selerror("UnsupportedConversion");
						 SELSTDTHROW;
						 break;
	}
	delete d;

	return value;
} // intColData()


ASCregionTable::ASCregionTable(const std::string &setName,
							   const std::string &xCoordName,
							   const std::string &yCoordName,
							   bool closeSet = true)
			  : _setName(setName), _xCoord(xCoordName), _yCoord(yCoordName),
				_closeSet(closeSet)
{
} // ASCregionTable::ASCregionTable()


void InASCregionTable::initFromTable()
{
	if (_table->attribute("HDUCLAS1")->asString() != "REGION") {
		selerrstr<<"The table in file `"<<_setName<<"' is not of type "
				   "`REGION' (HDUCLASS1="
				 <<_table->attribute("HDUCLAS1")->asString()<<")"
				 <<selerror("UnknownRegionTableType");
		SELSTDTHROW;
	}

	std::string posColNames = _table->attribute("MFORM1")->asString();
	std::string::size_type pos = posColNames.find(",");
	if (pos != std::string::npos) {
		_xCoord = posColNames.substr(0, pos);
		_yCoord = posColNames.substr(pos+1, posColNames.length()-pos-1);

		if (!_table->hasColumn(_xCoord) || !_table->hasColumn(_yCoord)) {
			selerrstr<<"Region table "
					   "in file `"<<_setName<<"' does not contain "
					   "columns with names `"<<_xCoord<<"' and `"
					  <<_yCoord<<"' as indicated in the MFORM1 attribute"
					  <<selerror("ColumnNotFound", 2);
			SELSTDTHROW;
		}
	} else {
		selerrstr<<"Cannot figure names of coordinate axes from MFORM1 "
				   "attribute value `"<<posColNames<<"'"
				 <<selerror("UnknownCoordinateAxes", 3);
		SELSTDTHROW;
	}

	_columnNames.push_back("SHAPE");	// shape name
	_columnNames.push_back(_xCoord);	// x-coordinates
	_columnNames.push_back(_yCoord);	// y-coordinates
	_columnNames.push_back("R");		// radius
	_columnNames.push_back("ROTANG");	// rotation angle
	_columnNames.push_back("COMPONENT");// component number

	for (unsigned t = SHAPE; t<=COMPONENT; ++t) {
		_column.push_back(static_cast<Column *>(0));
		if (_table->hasColumn(_columnNames[t]))
			_column[t] = _table->column(_columnNames[t]);
	}

	//	fill component map
	if (_column[COMPONENT]) {
		unsigned compIndx = 0;
		for (unsigned n = 0; n<_table->rows(); ++n) {
			unsigned compNr = intColData(_column[COMPONENT], n);
			if (_compMap.find(compNr) == _compMap.end()) {
				MapType::value_type p(compNr, compIndx++);
				_compMap.insert(p);
			}
		}
	} else {
		MapType::value_type p(1, 1);
		_compMap.insert(p);
	}
} // InASCregion::initFromTable()


InASCregionTable::InASCregionTable(const std::string &setName)
					: ASCregionTable(setName, "X", "Y", true)
{
	SetTableColumn stc(setName);
	_set	= stc.dataSet();
	_table	= stc.table();

	initFromTable();
} // InASCregionTable::InASCregionTable()


InASCregionTable::InASCregionTable(DataSet *dataSet)
{
	_set		= dataSet;
	_setName	= _set->name();
	_closeSet	= false;
	_table		= _set->table("REGION");
	initFromTable();
} // InASCregionTable::InASCregionTable()


InASCregionTable::InASCregionTable(Table *table)
{
	_set		= table->parent();
	_setName	= _set->name();
	_closeSet	= false;
	_table		= table;
	initFromTable();
} // InASCregionTable::InASCregionTable()


InASCregionTable::~InASCregionTable()
{
	if (_closeSet)
		dataSetServer->close(_set);
} // InASCregionTable::~InASCregionTable()



unsigned InASCregionTable::elements() const
{
	return _table->rows();
} // InASCregionTable::elements()


unsigned InASCregionTable::components() const
{
	return _compMap.size();
} // InASCregionTable::components()


unsigned InASCregionTable::component(unsigned elementNr)
{
	if (_column[COMPONENT]) {
		unsigned compNr = intColData(_column[COMPONENT], elementNr);
		return _compMap[compNr];
	}

	return _compMap[0];
} // InASCregionTable::component()


bool InASCregionTable::included(unsigned elementNr) const
{
	if (!_column[SHAPE])
		return true;	// element is selected by default

	ColumnData *d = _column[SHAPE]->data();
	bool b = (d->stringData()+d->position(elementNr))[0] != '!';
	delete d;

	return b;
} // InASCregionTable::included()


ShapeBasics::ShapeType InASCregionTable::whatShape(unsigned elementNr) const
{
	if (!_column[SHAPE])
		return ShapeBasics::POINT;

	ColumnData *d = _column[SHAPE]->data();
	std::string shape(d->stringData() + d->position(elementNr));
	delete d;
	if (shape[0] == '!')	//	region is excluded
		shape = shape.substr(1, shape.length()-1);
	shape = uppercase(shape);

	if (shape == "POINT")				return ShapeBasics::POINT;
	else if (shape == "CIRCLE")			return ShapeBasics::CIRCLE;
	else if (shape == "ANNULUS")		return ShapeBasics::ANNULUS;
	else if (shape == "ELLIPSE")		return ShapeBasics::ELLIPSE;
	else if (shape == "ELLIPTANNULUS")	return ShapeBasics::ELLIPTANNULUS;
	else if (shape == "RECTANGLE")		return ShapeBasics::RECTANGLE;
	else if (shape == "ROTRECTANGLE")	return ShapeBasics::ROTRECTANGLE;
	else if (shape == "BOX")			return ShapeBasics::BOX;
	else if (shape == "ROTBOX")			return ShapeBasics::ROTBOX;
	else if (shape == "DIAMOND" || shape == "RHOMBUS")
										return ShapeBasics::RHOMBUS;
	else if (shape == "ROTDIAMOND" || shape == "ROTRHOMBUS")
										return ShapeBasics::ROTRHOMBUS;
	else if (shape == "SECTOR" || shape == "PIE")
										return ShapeBasics::SECTOR;
	else if (shape == "POLYGON")		return ShapeBasics::POLYGON;
	else {
		selerrstr<<"Unknown region shape `"<<shape<<"' in column "
				 <<_column[SHAPE]->qualifiedName()
				 <<selerror("UnknownShape");
		SELSTDTHROW;
	}

	return ShapeBasics::POINT;
} // ASCregionTable::ShapeType ASCregion::whatShape()


std::vector<float> InASCregionTable::data(DataType dataType, unsigned elementNr)
const
{
	if (!_column[dataType])
		return std::vector<float>();	//	no such column

	ColumnData	*d = _column[dataType]->data();
	std::vector<float>	v(d->elements());

	for (unsigned element=0; element<d->elements(); ++element) {
		unsigned long pos = d->position(elementNr, element);
		switch (_column[dataType]->dataType()) {
			case Dal::Int8:	  v[element] = d->int8Data()[pos]; break;
			case Dal::Int16:  v[element] = d->int16Data()[pos]; break;
			case Dal::Int32:  v[element] = d->int32Data()[pos]; break;
			case Dal::Real32: v[element] = d->real32Data()[pos]; break;
			case Dal::Real64: v[element] = d->real64Data()[pos]; break;
			default:
				selerrstr<<"Conversion error in data access to region table"
						 <<selerror("ConversionError");
				SELSTDTHROW;
				break;
		}
	}

	delete d;
	return v;
} // InASCregionTable::data()


//
//	OutASCregionTable
//
OutASCregionTable::OutASCregionTable(const std::string &setName,
									 const std::string &xCoordName = "X",
									 const std::string &yCoordName = "Y")
				: ASCregionTable(setName, xCoordName, yCoordName, true)
{
	_set = dataSetServer->open(setName, DataSet::Create);
	_table = 0;
} // OutASCregionTable::OutASCregionTable()


OutASCregionTable::OutASCregionTable(DataSet *dataSet,
									 const std::string &xCoordName = "X",
									 const std::string &yCoordName = "Y")
				: ASCregionTable("", xCoordName, yCoordName, false)
{
	_set		= dataSet;
	_setName	= _set->name();
	_table		= 0;
} // OutASCregionTable::OutASCregionTable()


OutASCregionTable::OutASCregionTable(Table *table,
									 const std::string &xCoordName = "X",
									 const std::string &yCoordName = "Y")
				: ASCregionTable("", xCoordName, yCoordName, false)
{
	_set		= table->parent();
	_setName	= _set->name();
	_table		= table;
} // OutASCregionTable::OutASCregionTable()


OutASCregionTable::~OutASCregionTable()
{
	if (_closeSet)
		dataSetServer->close(_set);
} // OutASCregionTable::~OutASCregionTable()


//
//	input/output operators <</>>
//
InASCregionTable &operator>>(InASCregionTable &regionTable, GeoRegion &region)
{
	for (unsigned n=0; n<regionTable.elements(); ++n) {
		//	get x, y coordinates
		std::vector<float> xCoord = regionTable.data(ASCregionTable::X, n),
					  yCoord = regionTable.data(ASCregionTable::Y, n);
		if (!xCoord.size() || !yCoord.size()) {
			selerrstr<<"Region table does not contain mandatory spatial - "
					   "columns X/Y - cannot continue"
					 <<selerror("MandatoryColumnsMissing");
					 SELSTDTHROW;
		}

		std::vector<float> radii = regionTable.data(ASCregionTable::R, n);
		std::vector<float> rotAngles =
								regionTable.data(ASCregionTable::ROTANG, n);
		//	shape reference point
		Point<float>	refPoint(xCoord[0], yCoord[0]);
		//	what component does the element belong to and is is in/excluded?
		unsigned compNr = regionTable.component(n);
		if (compNr >= region.size())
			// need to make more space in region for more components
			for (unsigned comp=0; comp<compNr-region.size()+1; ++comp)
				region.push_back(Component());

		bool included = regionTable.included(n);
		switch (regionTable.whatShape(n)) {

			case ShapeBasics::POINT:
				region[compNr].push_back(
						new FShape(refPoint, included));
				break;

			case ShapeBasics::CIRCLE: {
				if (radii.size() == 0) {
					selerrstr<<"Region table does not contain radius - "
							   "column R - cannot construct a circle"
							 <<selerror("ColumnMissing");
					SELSTDTHROW;
				}
				region[compNr].push_back(
						new FCircle(refPoint, radii[0], included));
				break;
			}

			case ShapeBasics::ANNULUS: {
				if (radii.size() <= 0) {
					selerrstr<<"Region table does not contain radius column "
					   "R of big enough size - cannot construct an "
					   "annulus"<<selerror("MissingData");
					 SELSTDTHROW;
				}
				region[compNr].push_back(
						new FAnnulus(refPoint, radii[0], radii[1], included));
				break;
			}

			case ShapeBasics::ELLIPSE:
				if (radii.size() < 2 || rotAngles.size() < 1) {
					selerrstr<<"Region table does not contain radius/rotation"
							   " angle column R/ROTANG of big enough size - "
							   "cannot construct an ellipse"
							 <<selerror("MissingData");
					SELSTDTHROW;
				}
				region[compNr].push_back(
						new FEllipse(refPoint, radii[0], radii[1],
									 rotAngles[0], included));
				break;

			case ShapeBasics::ELLIPTANNULUS:
				assert(radii.size() >= 4 && rotAngles.size() >= 2);
				region[compNr].push_back(
						new FElliptAnnulus(refPoint, radii[0], radii[1],
										   radii[2], radii[3], rotAngles[0],
										   rotAngles[1], included));
				break;

			case ShapeBasics::SECTOR:
				assert(rotAngles.size() >= 2);
				region[compNr].push_back(
						new FSector(refPoint, rotAngles[0], rotAngles[1],
									included));
				break;

			case ShapeBasics::RECTANGLE:
				assert(xCoord.size() >= 2 && yCoord.size() >= 2);
				region[compNr].push_back(
						new FRectangle(refPoint,
									   FPoint(xCoord[1], yCoord[1]),
									   included));
				break;

			case ShapeBasics::ROTRECTANGLE:
				assert(xCoord.size() >= 2 && yCoord.size() >= 2);
				assert(rotAngles.size() >= 1);
				region[compNr].push_back(
						new FRotRectangle(refPoint,
										  FPoint(xCoord[1], yCoord[1]),
										  rotAngles[0], included));
				break;

			case ShapeBasics::BOX:
				assert(radii.size() >= 2);
				region[compNr].push_back(
						new FBox(refPoint, radii[0], radii[1], included));
				break;

			case ShapeBasics::ROTBOX:
				assert(radii.size() >= 2 && rotAngles.size() >= 1);
				region[compNr].push_back(
						new FRotBox(refPoint, radii[0], radii[1],
									rotAngles[0], included));
				break;

			case ShapeBasics::RHOMBUS:
				assert(radii.size() >= 2);
				region[compNr].push_back(
						new FRhombus(refPoint, radii[0], radii[1], included));
				break;

			case ShapeBasics::ROTRHOMBUS:
				assert(radii.size() >= 2 && rotAngles.size() >= 1);
				region[compNr].push_back(
						new FRotRhombus(refPoint, radii[0], radii[1],
										rotAngles[0], included));
				break;

			case ShapeBasics::POLYGON: {
				FShape *s =
						new FPolygon(xCoord, yCoord, xCoord.size(), included);
				region[compNr].push_back(s);

				//	have to assign first to s, otherwise g++ bails
				//	out with weird error message
				break;
			}

			default:
				selerrstr<<"Unknown region type: "
						  <<static_cast<unsigned>(regionTable.whatShape(n))
						  <<" in output of region table - "
						  <<" will not be considered"
						  <<selwarning("UnknonShape");
				break;
		}
	}
	return regionTable;
} // operator>>


template<class T>
void vectorCol(const Column *col, unsigned rowNr, const std::vector<T> &v)
{
	const ColumnData *d = col->data();
	real32 *r32 = d->real32Data();
	for (unsigned element=0; element < d->elements(); ++element) {
		unsigned long pos = d->position(rowNr, element);
		r32[pos] = element < v.size()
					? static_cast<float>(v[element])
					: 0.f;
	}
	delete d;
} // vectorCol


OutASCregionTable &operator<<(OutASCregionTable &regionTable,
							  const GeoRegion &region)
{
	const unsigned long lengthShapeCol = 16;

	if (!regionTable._table)
		regionTable._table = regionTable._set->addTable("REGION",
													region.elements());
	*regionTable._table->addAttribute("HDUCLASS", "", "") = "ASC";
	*regionTable._table->addAttribute("HDUCLAS1", "", "") = "REGION";
	*regionTable._table->addAttribute("HDUCLAS2", "", "") = "STANDARD";
	*regionTable._table->addAttribute("MTYPE1", "", "") = "pos";
	*regionTable._table->addAttribute("MFORM1", "", "") =
				regionTable._xCoord + "," + regionTable._yCoord;

	unsigned maxPoints = region.maxPoints(),
			 maxRadii = region.maxRadii(),
			 maxRotAngles = region.maxRotAngles();

	std::vector<unsigned long> shapeSize;
	shapeSize.push_back(lengthShapeCol);
	Column *shapeCol = regionTable._table->addColumn("SHAPE", Dal::DString,
								"shape of element", "", shapeSize);
	std::vector<unsigned long> coordSize;
	coordSize.push_back(region.hasPolygon()
									? maxPoints + 1 // for terminating vertex
									: maxPoints);

	Column *xCol = regionTable._table->addColumn(regionTable._xCoord,
							Dal::Real32, "x-coordinate vector", "",coordSize),
		   *yCol = regionTable._table->addColumn(regionTable._yCoord,
							Dal::Real32, "y-coordinate vector", "", coordSize);

	Column *rCol = 0, *angCol = 0;
	if (maxRadii) {
		std::vector<unsigned long> rSize; rSize.push_back(maxRadii);
		rCol = regionTable._table->addColumn("R", Dal::Real32,
										 "radius vector", "", rSize);
	}

	if (maxRotAngles) {
		std::vector<unsigned long> angSize; angSize.push_back(maxRotAngles);
		angCol = regionTable._table->addColumn("ROTANG", Dal::Real32,
										 "rotation angle vector", "", angSize);
	}

	Column *compCol = regionTable._table->addColumn("COMPONENT", Dal::Int8,
								"component number that shape belongs to");

	//	loop over all elements
	unsigned elementNr = 0;
	ColumnData *sd = shapeCol->data(),
			   *cd = compCol->data();
	for (unsigned comp=0; comp<region.size(); ++comp)
		for (unsigned element=0; element<region[comp].size(); ++element) {
			char *s = sd->stringData() + sd->position(elementNr);
			std::string shapeName =
				std::string(region[comp][element]->included() ? "" : "!") +
					   region[comp][element]->name();
			strncpy(s, shapeName.c_str(), lengthShapeCol+1);

			std::vector<float> xPoints(region[comp][element]->pointsX()),
						  yPoints(region[comp][element]->pointsY());
			if (region[comp][element]->type() == ShapeBasics::POLYGON) {
				//	add terminal vertex
				xPoints.push_back(xPoints[0]);
				yPoints.push_back(yPoints[0]);
			}
			vectorCol(xCol, elementNr, xPoints);
			vectorCol(yCol, elementNr, yPoints);
			if (rCol)
				vectorCol(rCol, elementNr, region[comp][element]->radii());
			if (angCol)
				vectorCol(angCol, elementNr,
										region[comp][element]->rotAngles());
			cd->int8Data()[elementNr] = comp;
			++elementNr;
		}

	delete sd;
	delete cd;

	return regionTable;
} // operator<<
