//----------------------------------------------------------------------------
//
//                      XMM Science Analysis System (SAS)
//                     (c) 1997-2010 European Space Agency
//
//---------------------------------------------------------------------------
//
//	ExprTranslator.cc: --- implementation of ExprTranslator class
//
//	DESCRIPTION:
//		This implements the ExprTranlator class
//
//	CHANGE HISTORY:
//		UL, 1999-02-17: 1.0.0: - first version
//		UL, 1999-02-17: 1.1.0: - generated expression was invalid ;)
//		UL, 1999-09-20: 1.2.0: - take binning factors into account
//		UL, 1999-09-21: 1.3.0: - take also image offset into account
//		UL, 1999-10-25: 1.4.0: - added support for 2D regions from dsplot
//		UL, 1999-12-15: 1.4.1: - closed two minor memory leaks
//
//	AUTHOR:
//		U. Lammers (UL), Wed Feb 17 13:22:25 MET 1999
//
//	$Header: /data/xmm02/cvsadmin/soc_repos/sas/packages/selectlib/src/utils/ExprTranslator.cc,v 1.2 2001/08/10 16:15:00 ulammers Exp $
//

#include <iomanip>
#include <algorithm>
#include <strstream>

#include "common.h"
#include "stringutil.h"

#include "SelectError.h"
#include "ExprTranslator.h"


std::string ImgdisplayRegionSpecification::toSelectlibExpression(
											const std::string &xcolName,
											const std::string &ycolName,
											double binFactorX,
											double binFactorY,
											double minX,
											double minY) const
{
	std::string region(_spec);	// make a copy
	char exclIncl = region[0];
	if (exclIncl != '+' && exclIncl != '-')
		throw InvalidRegionSpec();

	std::string::size_type pos;
	if ((pos = region.find("#")) != std::string::npos)
		region=trim(region.substr(0, pos-1));

	pos = region.find('(');
	std::string regionType(lowercase(region.substr(1, pos-1))),
		   regionArgs(region.substr(pos+1, region.find_last_of(')')-pos-1));

	if (regionType == "circle") {
		std::string rotAng = "";
		if (binFactorX != binFactorY) {
			regionType = "ellipse";	//	a circle becomes an ellipse for
									//	different bin factors
			rotAng = ",0";			//	0: rotation angle

		}
		regionArgs = circleArgs(regionArgs, binFactorX, binFactorY,
								minX, minY) + rotAng;
	} else if (regionType == "ellipse")
		regionArgs = ellipseArgs(regionArgs, binFactorX, binFactorY,
							     minX, minY);
	else if (regionType == "box")
		regionArgs = boxArgs(regionArgs, binFactorX, binFactorY, minX, minY);
	else if (regionType == "polygon")
		regionArgs = polygonArgs(regionArgs, binFactorX, binFactorY,
								 minX, minY);
	else if (regionType == "point")
		regionArgs = pointArgs(regionArgs, binFactorX, binFactorY, minX, minY);
	else
		errstr<<"`"<<regionType<<"' is not a supported region"
			  <<Sas::error("UnsupportedRegion");

	std::string expr('(' + xcolName + ',' + ycolName + ") IN " +
				regionType + '(' + regionArgs + ')');

	return exclIncl == '+' ? expr : "!" + expr;
} // ImgdisplayRegionSpecification::toSelectlibExpression()


std::string ImgdisplayRegionSpecification::circleArgs(const std::string &s,
					 double binX, double binY, double minX, double minY) const
{
	std::vector<std::string> v(breakup(s, ","));
	if (v.size() != 3)
		errstr<<"Arguments of circle filter must be <x>,<y>,<r> but there "
				"is "<<s<<Sas::error("InputError");

	if (binX != binY)
		return convertPosition(v[0], binX, minX, v[1], binY, minY) + ',' +
			   convertDimension(v[2], binX) + ',' +
			   convertDimension(v[2], binY);

	std::string ret(convertPosition(v[0], binX, minX, v[1], binY, minY) + ',' +
			   convertDimension(v[2], binX));
	return ret;
} // ImgdisplayRegionSpecification::circleArgs()


std::string ImgdisplayRegionSpecification::ellipseArgs(const std::string &s,
					 double binX, double binY, double minX, double minY) const
{
	std::vector<std::string> v(breakup(s, ","));
	if (v.size() != 5)
		errstr<<"Arguments of ellipse filter must be <x>,<y>,<r1>,<r2>, "
				"<ANGLE> but there is "<<s<<Sas::error("InputError");

	std::string ret(convertPosition(v[0], binX, minX, v[1], binY, minY) + ',' +
			   convertDimension(v[2], binX) + ',' +
			   convertDimension(v[3], binY) + ',' +
			   v[4]);
	return ret;
} // ImgdisplayRegionSpecification::ellipseArgs()


std::string ImgdisplayRegionSpecification::boxArgs(const std::string &s,
					 double binX, double binY, double minX, double minY) const
{
	std::vector<std::string> v(breakup(s, ","));
	if (v.size() != 5)
		errstr<<"Arguments of box filter must be <x>,<y>,<w>,<y>,<angle> "
				"but there is "<<s<<Sas::error("InputError");

	
	std::string ret(convertPosition(v[0], binX, minX, v[1], binY, minY) + ',' +
			   convertDimension(v[2], .5 * binX) + ',' +
			   convertDimension(v[3], .5 * binY) + ',' +
			   v[4]);			//	selectlib box filter expects half-widths
	return ret;
} // ImgdisplayRegionSpecification::boxArgs()


std::string ImgdisplayRegionSpecification::polygonArgs(const std::string &s,
					 double binX, double binY, double minX, double minY) const
{
	std::vector<std::string> v(breakup(s, ","));
	unsigned N = v.size();
	if (N % 2)
		errstr<<"Arguments of box filter must be <x1>,<y1>,<x2>,<y2>, "
				"..., <xn>,<yn> but there is "<<s<<Sas::error("InputError");

	std::string res;
	for (unsigned n=0; n<N/2; ++n)
		res += convertPosition(v[2*n], binX, minX, v[2*n+1], binY, minY) + ',';

	return res.substr(0, res.length()-1);
} // ImgdisplayRegionSpecification::polygonArgs()


std::string ImgdisplayRegionSpecification::pointArgs(const std::string &s,
					 double binX, double binY, double minX, double minY) const
{
	std::vector<std::string> v(breakup(s, ","));
	if (v.size() != 2)
		errstr<<"Arguments of point filter must be <x>,<y> but"
				" there is "<<s<<Sas::error("InputError");

	return convertPosition(v[0], binX, minX, v[1], binY, minY);
} // ImgdisplayRegionSpecification::pointArgs()


std::string ImgdisplayRegionSpecification::convertPosition(
													  const std::string &x,
													  double binX, double minX,
													  const std::string &y,
													  double binY, double minY)
const
{
	double xv = toNumber<double>(x),
		   yv = toNumber<double>(y);

	return toString((xv-1.) * binX + minX) + ',' +
		   toString((yv-1.) * binY + minY);
} // ImgdisplayRegionSpecification::convertPosition()


std::string ImgdisplayRegionSpecification::convertDimension(
													   const std::string &dim,
													   double f) const
{
	return toString(toNumber<double>(dim) * f);
} // ImgdisplayRegionSpecification::convertDimension()


DsplotRegionSpecification::RegionType
DsplotRegionSpecification::type(const std::vector<std::string> &comp) const
{
	static char *_s[] = {"above", "below", "left", "right", "polyi", "polyo" };
	static std::vector<std::string> s(_s, _s + NUM_OF(_s));

	RegionType reg = NONE;
	bool on = false;
	for (unsigned n=0; n<comp.size(); ++n) {
		std::vector<std::string> v(breakup(comp[n], " ,"));
		if (v.size() < 1)
			continue;
		if (v[0] == "type") {
			std::vector<std::string>::const_iterator i =
											::find(s.begin(), s.end(), v[1]);
			if (i != s.end())
				reg = static_cast<RegionType>(i-s.begin());
			else
				throw InvalidRegionSpec();
		} else if (uppercase(v[0]) == "ON")
			on = true;
	}

	return on && reg != NONE ? reg : NONE;
} // DsplotRegionSpecification::type()

std::string DsplotRegionSpecification::toSelectlibExpression(
												const std::string &xColName,
												const std::string &yColName,
					 							double, double,
												double, double) const
{
	std::vector<std::string> comp(breakup(_spec, ";"));
	if (comp.size() < 3)
		throw InvalidRegionSpec();

	RegionType whatType = type(comp);
	if (whatType == NONE)
		return "";

	//
	//	we have a polygon
	//
	std::vector<double> x, y;
	if (whatType == INSIDE_POLY || whatType == OUTSIDE_POLY) {
		for (unsigned n=1; n<comp.size(); ++n) {
			std::vector<std::string> s(breakup(comp[n], " ,"));
			if (s.size() != 3)
				continue;
			if (s[0] != "xy")
				continue;
			x.push_back(toNumber<double>(s[1]));
			y.push_back(toNumber<double>(s[2]));
		}

		if (x.size() < 3)
			//	cannot have a polygon with less than 3 points
			throw InvalidRegionSpec();
	} else
		return "";

	std::strstream s;
	if (whatType == OUTSIDE_POLY)
		//	outside polygon
		s<<'!';	
	s<<'('<<xColName<<","<<yColName<<") IN polygon(";
	unsigned N = x.size();
	for (unsigned n=0; n<N; ++n) {
		s<<std::setprecision(6)<<x[n]<<","<<std::setprecision(6)<<y[n];
		if (n != N-1)
			s<<',';
	}
	s<<')'<<std::ends;

	char *p = s.str();
	std::string ret(p);
	delete p;

	return ret;
} // DsplotRegionSpecification::toSelectlibExpression()


std::string DsplotRegionSpecification::toSelectlibExpression(
										const std::string &colName) const
{
	std::vector<std::string> comp(breakup(_spec, ";"));
	if (comp.size() < 3)
		throw InvalidRegionSpec();

	RegionType whatType = type(comp);

	unsigned n;
	double x1 = 0.,
		   y1 = 0.,
		   x2 = 0.,
		   y2 = 0.;
	for (n=0; n<comp.size(); ++n) {
		std::vector<std::string> v(breakup(comp[n], " ,"));
		if (v[0] == "line") {
			//	found it
			x1 = toNumber<double>(v[1]);
			y1 = toNumber<double>(v[2]);
			x2 = toNumber<double>(v[3]);
			y2 = toNumber<double>(v[4]);
			break;
		}
	}

	if (n >= comp.size())
		throw InvalidRegionSpec();

	if (whatType != NONE) {
		//
		//	sometimes you get "line 0, 0, 0, 0"
		//

		if (whatType == ABOVE_LINE || whatType == BELOW_LINE) {
			//
			//	a single range
			//
			std::strstream s1, s2;
			s1<<std::setprecision(6)<<(x1 < x2 ? x1 : x2)<<std::ends;
			s2<<std::setprecision(6)<<(x1 < x2 ? x2 : x1)<<std::ends;
			char *s1p = s1.str(), *s2p = s2.str();
			std::string s1str(s1p), s2str(s2p);
			delete s1p; delete s2p;

			return colName + " IN [" + s1str + ":" + s2str + ']';
		} else if (whatType == LEFT_OF_LINE || whatType == RIGHT_OF_LINE) {
			//
			//	everything to the left or right of the specified line
			//	take the mean of x1/x2 as the value
			//
			std::strstream s;
			s<<std::setprecision(6)<<.5*(x1+x2)<<std::ends;
			char *p = s.str();
			std::string str(p);
			delete p;

			return colName + (whatType == LEFT_OF_LINE ? "<=" : ">=") + str;
		} else
			throw InvalidType();
	}

	return "";
} // DsplotRegionSpecification::toSelectlibExpression()


