//----------------------------------------------------------------------------
//
//                     XMM Science Analysis System (SAS)
//                    (c) 1997-2010 European Space Agency
//
//---------------------------------------------------------------------------
//
//	OGIPgti.cc: --- implementation of OGIPgtiTable classes
//
//	DESCRIPTION:
//		Instances of class OGIPgtiTable represent GTI filters
//		according to OGIP conventions
//
//	CHANGE HISTORY:
//		UL, 1998-03-25: 1.0.0: - first version
//		UL, 1998-03-03: 1.1.0: - modifications enforced by Lowmem Dal
//		UL, 2000-05-22: 1.2.0: - bug fix
//		UL, 2000-05-25: 1.3.0: - add HDUCLAS[S|1|2] keywords to GTI tables
//								 to make them as being OGIP compliant
//							   - OutOGIPgtiTable can now be constructed
//								 also from existing tables, e.g.
//									OutOGIPgtiTable gti("gti.ds")
//									OutOGIPgtiTable gti("gti.ds:STDGTI")
//		UL, 2000-06-19: 1.4.0: - fixed bug in operator>> which caused
//								 overwriting of GTI tables to fail if the
//								 new table is smaller than the old one
//		UL, 2002-01-29: 1.5.0: - added OutOGIPgtiTable::mergeTimingAttributes()
//								 fixes SCR 58: propagate timing related
//								 keywords in GTI merging
//		UL, 2002-01-31: 1.5.1: - bug fix in new mergeTimingAttributes(): timing
//								 keywords were not treated as optional
//		UL, 2002-04-08: 1.5.2: - allow timing attributes to be of types
//								 real/int or int/real for equality check
//								 such that 0.0 = 0 yields true
//
//	AUTHOR:
//		U. Lammers (UL), Wed Mar 25 16:52:23 MET 1998
//
//	$Header: /usr/local/cvs/sasdev/sas/packages/selectlib/src/utils/OGIPgti.cc,v 1.6 2002/04/08 15:45:44 ulammers Exp $
//
#include <algorithm>

#include "OGIPgti.h"
#include "errstr.h"
#include "common.h"
#include "SetTableColumn.h"
#include "Tracer.h"

OGIPgtiTable::OGIPgtiTable(const std::string &setName, bool closeSet = true)
			: _setName(setName), _closeSet(closeSet)
{
} // OGIPgtiTable::OGIPgtiTable()


InOGIPgtiTable::InOGIPgtiTable(const std::string &setName)
					: OGIPgtiTable(setName, true)
{
	SetTableColumn stc(setName);
	_set	= stc.dataSet();
	_table	= stc.table();
} // InOGIPgtiTable::InOGIPgtiTable()


InOGIPgtiTable::InOGIPgtiTable(DataSet *dataSet)
{
	_set		= dataSet;
	_setName	= _set->name();
	_closeSet	= false;
	_table		= _set->table("STDGTI");
} // InOGIPgtiTable::InOGIPgtiTable()


//	construct from table
InOGIPgtiTable::InOGIPgtiTable(Table *table)
{
	_set		= table->parent();
	_setName	= _set->name();
	_closeSet	= false;
	_table		= table;
} // InOGIPgtiTable::InOGIPgtiTable()


InOGIPgtiTable::~InOGIPgtiTable()
{
	if (_closeSet)
		dataSetServer->close(_set);
} // InOGIPgtiTable::~InOGIPgtiTable


InOGIPgtiTable &operator>>(InOGIPgtiTable &gtiTable, GtiList &gtiList)
{
	ColumnData *startData = gtiTable._table->column(GTISTARTCOLNAME)->data(),
			   *stopData = gtiTable._table->column(GTISTOPCOLNAME)->data();
	real64 *start	= startData->real64Data(),
		   *stop	= stopData->real64Data();

	//	read the GTIs
	for (unsigned i=0; i<gtiTable._table->rows(); ++i) {
		Gti gti(start[i], stop[i]);
		gtiList.push_back(gti);
	}
	delete startData;
	delete stopData;

	//	sort the vector
	gtiList.sortIt();

	gtiList.addOrigin(gtiTable._table->qualifiedName());

	return gtiTable;
} // operator>>


namespace {
	bool operator==(const Attribute &att1, const Attribute &att2) {
		if (att1.dataType() == att2.dataType()) 
			switch (att1.dataType()) {
				case Attribute::Bool:
					return att1.asBool() == att2.asBool(); break;
				case Attribute::Int:
					return att1.asInt() == att2.asInt(); break;
				case Attribute::Real:
					return att1.asReal() == att2.asReal(); break;
				case Attribute::DString:
					return att1.asString() == att2.asString(); break;
				default: break;
			}

		//	also allow the combo int/real + real/int
		if ((att1.dataType() == Attribute::Real &&
										att2.dataType() == Attribute::Int) ||
			(att2.dataType() == Attribute::Real &&
										att1.dataType() == Attribute::Int))
			return att1.asReal() == att2.asReal();

		return false;
	}
	bool operator!=(const Attribute &att1, const Attribute &att2) {
		return !operator==(att1, att2);
	}
}


OutOGIPgtiTable::OutOGIPgtiTable(const std::string &setName)
				: OGIPgtiTable(setName, true)
{
	std::vector<std::string> v = SetBlock::componentNames(setName);
	assert(v.size() == 2);
	_set = dataSetServer->exists(v[0])
				? dataSetServer->open(v[0], DataSet::Modify)
				: dataSetServer->open(v[0], DataSet::Create);

	setFromSet(_set, v[1]);
} // OutOGIPgtiTable::OutOGIPgtiTable()


void OutOGIPgtiTable::setFromSet(DataSet *set, const std::string &tableName)
{
	_set = set;
	std::string gtiTabName = tableName.length() ? tableName : "STDGTI";
	_table = _set->hasBlock(gtiTabName)
				? _set->table(gtiTabName)
				: _set->addTable(gtiTabName, 0);
	setFromTable(_table);
} // OutOGIPgtiTable::setFromSet()


void OutOGIPgtiTable::setFromTable(Table *table)
{
	_table = table;

	setAttribute("HDUCLASS", "format conforms to OGIP standard", "OGIP");
	setAttribute("HDUCLAS1", "table contains Good Time Intervals","GTI");
	setAttribute("HDUCLAS2", "standard Good Time Interval table",\
					   "STANDARD");

	std::vector<unsigned long> s; s.push_back(1);
	if (!_table->hasColumn(GTISTARTCOLNAME))
		_table->addColumn(GTISTARTCOLNAME, Data::Real64,
						  "lower GTI boundary", "s", s);
	if (!_table->hasColumn(GTISTOPCOLNAME))
		_table->addColumn(GTISTOPCOLNAME, Data::Real64,
						  "upper GTI boundary", "s", s);
} // OutOGIPgtiTable::setFromTable()


OutOGIPgtiTable::OutOGIPgtiTable(DataSet *dataSet)
				: OGIPgtiTable(dataSet->name(), false)
{
	setFromSet(dataSet);
} // OutOGIPgtiTable::OutOGIPgtiTable()


OutOGIPgtiTable::OutOGIPgtiTable(Table *table)
				: OGIPgtiTable(table->parent()->name(), false)
{
	setFromTable(table);
} // OutOGIPgtiTable::OutOGIPgtiTable()


void OutOGIPgtiTable::setAttribute(const std::string &name,
								   const std::string &comment,
								   const std::string &value)
{
	Attribute *att;
	if (_table->hasAttribute(name))
		att = _table->attribute(name);
	else {
		att = _table->addAttribute(name, comment);
		*att = value;
	}

	if (att->asString() != value) {
		errstr<<"Changing value of `"<<att->qualifiedName()<<"' from `"
			  <<*att<<"' to `"<<value<<"'"
			  <<Sas::warning("ChangingAttributeValue");
		*att = value;
	}
} // OutOGIPgtiTable::setAttribute()


namespace {
class DataSetExists {
public:
	bool operator()(const string &s) { return dataSetServer->exists(s); }
};

class SetClose {
public:
	void operator()(Table *tab) { dataSetServer->close(tab->parent()); }
};
}

//
//	check that all timing related attributes in the GtiList have
//	identical values and write those to the output table; if
//	none are found add attributes with d/f values
//
void OutOGIPgtiTable::mergeTimingAttributes(const GtiList &gtiList)
{
	const char *attribs[] = { "TIMEUNIT", "TIMESYS", "MJDREF", "TIMEREF",
							  "TASSIGN", "TIMEZERO", "CLOCKAPP" };
	vector<Attribute *>	attVec;
	vector<Table *>		tabs;
	const vector<string> &tabNames = gtiList.origins();

	DataSetExists setExists;
	vector<string>::const_iterator i =
					find_if(tabNames.begin(), tabNames.end(), setExists);
	if (i != tabNames.end()) {
		tabs.push_back(dataSetServer->open(*i, Dal::Read,
											DataSetServer::Low)->table(*i));
		for (++i; i != tabNames.end(); ++i)
			if (setExists(*i)) {
				tabs.push_back(dataSetServer->open(*i, Dal::Read,
									DataSetServer::Low)->table(*i));
			}
		unsigned ntabs = tabs.size();
		for (unsigned n=0; n<NUM_OF(attribs); ++n) {
			string attname(attribs[n]);
			unsigned k;
			for (k=1; k<ntabs; ++k)
				if (tabs[0]->hasAttribute(attname) &&
					tabs[k]->hasAttribute(attname)) {
					if (*tabs[0]->attribute(attname) !=
						*tabs[k]->attribute(attname)) {
							errstr<<"Attribute `"<<attname<<"' has different "
									"value in `"
								  <<tabs[0]->qualifiedName()<<"' ("
								  <<tabs[0]->attribute(attname)->asString()
								  <<") and `"
								  <<tabs[k]->qualifiedName()<<"' ("
								  <<tabs[k]->attribute(attname)->asString()
								  <<") - please make sure it makes sense to "
									"merge the GTI tables"
								  <<Sas::warning("DifferentAttributeValues");
							break;
						}
				} else
					break;
			if (k>=ntabs)
				attVec.push_back(tabs[0]->attribute(attname));
		}
		//	write attributes to outtable
		unsigned N = attVec.size();
		for (unsigned n=0; n<N; ++n)
			_table->addAttribute(attVec[n]);
	} else {
		//	no pre-defined attribute values - provide defauls
		*_table->addAttribute(attribs[0], "All times in s unless specified "
				"otherwise") = "s";
		*_table->addAttribute(attribs[1], "XMM time will be TT (Terrestial "
				"Time)") = "TT";
		*_table->addAttribute(attribs[2], "1998-01-01T00:00:00 (TT) expressed "
				"in MJD") = 50814.;
		*_table->addAttribute(attribs[3], "Reference location of photon "
				"arrival times") = "LOCAL";
		*_table->addAttribute(attribs[4], "Location of time assignment") =
				"SATELLITE";
		*_table->addAttribute(attribs[5], "Clock correction (if not zero)") =0;
		*_table->addAttribute(attribs[6], "Clock correction applied?") = true;
	}

	for_each(tabs.begin(), tabs.end(), SetClose());
} // OutOGIPgtiTable::mergeTimingAttributes()


OutOGIPgtiTable::~OutOGIPgtiTable()
{
	if (_closeSet)
		dataSetServer->close(_set);
} // OutOGIPgtiTable::~OutOGIPgtiTable()


OutOGIPgtiTable &operator<<(OutOGIPgtiTable &table, const GtiList &gtiList)
{
	GtiList reducedList(gtiList);
	reducedList.reduce();

	unsigned nold = table._table->rows(),
			 nnew = reducedList.size();
	if (nnew > nold)
		//	need to add rows
		table._table->insertRows(nold, nnew - nold);
	else if (nnew < nold)
		//	need to delete rows
		table._table->deleteRows(nnew, nold - nnew);

	ColumnData *startData = table._table->column(GTISTARTCOLNAME)->data(),
			   *stopData = table._table->column(GTISTOPCOLNAME)->data();
	real64 *startcol = startData->real64Data(),
		   *stopcol	= stopData->real64Data();

	for (unsigned n=0; n<reducedList.size(); ++n) {
		startcol[n] = reducedList[n].start();
		stopcol[n] = reducedList[n].stop();
	}
	*table._table->addAttribute("ONTIME",
			"sum of all Good Time Intervals", "s") = reducedList.ontime();
	if (!reducedList.empty()) {
		*table._table->addAttribute("TSTART", "Lower bound of first GTI", "s")
			= reducedList.front().start();
		*table._table->addAttribute("TSTOP", "Uppler bound of last GTI", "s")
			= reducedList.back().stop();
	}
	
	table.mergeTimingAttributes(gtiList);

	delete startData;
	delete stopData;

	return table;
} // operator<<


