//----------------------------------------------------------------------------
//
//                     XMM Science Analysis System (SAS)
//                    (c) 1997-2010 European Space Agency
//
//---------------------------------------------------------------------------
//
//	SelectWCS.cc: --- implementation of SelectWCS class
//
//	DESCRIPTION:
//		Instances of class SelectWCS represent transformations to go
//		from integer pixel numbers to real-valued coordinates
//		(and vice-versa) according to World-Coordinate-System
//		Conventions.
//
//	CHANGE HISTORY:
//		UL, 1998-10-06: 1.0.0: - first version
//		UL, 1999-01-19: 1.1.0: - minor modification in SelectWCS::paramValue
//
//	REFERENCES:
//		[1] : E.W. Greisen, M. Calabretta, "Representation of Celestial
//			  Coordinates in FITS",
//			  ftp://fits.vc.nrao.edu/fits/documents/wcs/wcs_all.ps
//
//	AUTHOR:
//		U. Lammers (UL), Tue Oct  6 15:59:03 MET DST 1998
//
//	$Header: /data/xmm02/cvsadmin/soc_repos/sas/packages/selectlib/src/utils/SelectWCS.cc,v 1.2 2001/08/10 16:15:01 ulammers Exp $
//

#include "common.h"
#include "stringutil.h"
#include "SelectWCS.h"

double SelectWCS::paramValue(const Array *array, const std::string &keyName,
					   double dftValue = 0.)
{
	if (array->hasAttribute(keyName)) {
		_keywords = true;
		return array->attribute(keyName)->asReal();
	} else if (array->parent()->hasAttribute(keyName)) {
		//	is mask image is in primary and no "XMMSEP" keywords
		//	is present all attributes are assigned to the data set
		_keywords = true;
		return array->parent()->attribute(keyName)->asReal();
	}

	return dftValue;
} // SelectWCS::paramValue()


double SelectWCS::paramValue(const Table *table, const std::string &colName,
							 const std::string &attrName, double dftValue)
{
	const Column *col = table->column(colName);
	if (col)
		if (col->hasAttribute(attrName)) {
			_keywords = true;
			return col->attribute(attrName)->asReal();
		}

	return dftValue;
} // SelectWCS::paramValue()


std::string SelectWCS::SelectWCSattributeName(const std::string &stem,
											  unsigned short axisNr) const
{
	return stem + toString(axisNr);
} // WCS::WCSattributeName()


SelectWCS::SelectWCS(const Array *array, unsigned short axisNr)
{
	_keywords = false;
	if (array) {
		_refPixel = paramValue(array, SelectWCSattributeName("CRPIX", axisNr),
							   0.);
		_refPixelValue = paramValue(array, SelectWCSattributeName("CRVAL",
								axisNr), 0.);
		_delta = paramValue(array, SelectWCSattributeName("CDELT", axisNr),
								1.);
	} else {
		_refPixel = _refPixelValue = 0.;
		_delta = 1.;
	}
} // SelectWCS::SelectWCS()


SelectWCS::SelectWCS(const Table *table, const std::string &colName)
{
	_keywords = false;
	if (table) {
		_refPixel = paramValue(table, colName, "TCRPX", 0.);
		_refPixelValue = paramValue(table, colName, "TCRVL", 0.);
		_delta = paramValue(table, colName, "TCDLT", 1.);
	} else {
		_refPixel = _refPixelValue = 0.;
		_delta = 1.;
	}
} // SelectWCS::SelectWCS()


double SelectWCS::operator()(int pixel) const
{
	return (DBL(pixel)-_refPixel) * _delta + _refPixelValue;
} // SelectWCS::operator(int)


int SelectWCS::operator()(double value) const
{
	return INT((value - _refPixelValue)/_delta + _refPixel);
} // SelectWCS::operator(double)



std::ostream &operator<<(std::ostream &os, const SelectWCS &wcs)
{
	os<<"SelectWCS:\n"
	  <<"   reference pixel       : "<<wcs.refPixel()<<endl
	  <<"   reference pixel value : "<<wcs.refPixelValue()<<endl
	  <<"   delta                 : "<<wcs.delta()<<endl
	  <<"   keyword-based         : "<<
					(wcs.keywordsPresent() ? "true" : "false")<<endl;

	return os;
} // operator<<


