//
//	filtertest: test the Filter class
//
//	UL, Tue Mar 17 16:28:02 MET 1998
//
//	$Header: /data/xmm02/cvsadmin/soc_repos/sas/packages/selectlib/test/filtertestcxx.cc,v 1.8 2001/08/10 16:15:04 ulammers Exp $
//

#include <cassert>
#include <iostream>
#include <cmath>
#include <unistd.h>
#include "Dal.h"
#include "DataFilter.h"
#include "ExprColumnGenerator.h"

template<class type>
type SQR(type t) { return t*t; }

int main()
{
	const unsigned maxx = 256;
	const unsigned maxy = 256;
	const std::string qualColName = "FLAGS";

	//
	//	create a test set
	//
	DataSet *set = dataSetServer->open("filtertest.ds", DataSet::Create);
	Table *table = set->addTable("SPATIAL", maxx * maxy);
	Column *colx = table->addColumn("RAWX", Dal::Int16, "pixel", "raw X"),
		   *coly = table->addColumn("RAWY", Dal::Int16, "pixel", "rawY"),
		   *bcol = table->addColumn("BCOL", Dal::Bool, "", "RAWX + RAWY>100");

	ColumnData *cdx = colx->data(),
			   *cdy = coly->data(),
			   *cdb = bcol->data();
	int16 *x = cdx->int16Data(),
		  *y = cdy->int16Data();
	bool8 *b = cdb->boolData();

	for (unsigned i=1; i<=maxy; ++i)
		for (unsigned j=1; j<=maxx; ++j) {
			unsigned num = (i-1)*maxx+j-1;
			x[num] = j;
			y[num] = i;
			b[num] = (x[num]+y[num] > 100);
		}

	*table->addAttribute("RAWXSQ") = "(RAWX-1)**2";
	*table->addAttribute("RAWYSQ") = "(RAWY-1)**2";
	*table->addAttribute("DIST") = "sqrt(#RAWXSQ + #RAWYSQ)";
	*table->addAttribute("DISTSEL") = "#DIST <= 128.";
	*table->addAttribute("PRIMSRC") = "2";

	//	construct new column populated with the distance
	ExprColumnGenerator columnGenerator("sqrt((RAWX-1)**2 + (RAWY-1)**2)",
										"DIST");
	columnGenerator.doIt(*table, *table);

	//
	//	check against contents of boolean column
	//
	DataFilter filter01("RAWX+RAWY>100 == BCOL");
	assert(filter01.doSelect(*table, qualColName, 0) == table->rows());

	//
	//	select one row
	//
	DataFilter filter02("#PRIMSRC==#ROW && #E<#PI");
	assert(filter02.doSelect(*table, qualColName, 0) == 1);

	//
	//	select everything
	//
	DataFilter filter1("box(128.5,128.5,128,128,0,RAWX,RAWY)");
	assert(filter1.doSelect(*table, qualColName, 0) == table->rows());

	ColumnData *cdsel = table->column(qualColName)->data();
	int8 *sel = cdsel->int8Data();
	for (unsigned n=0; n<maxx * maxy; ++n)
		assert(sel[n] == 1);
	delete cdsel;
	std::cout<<"Filter1 test ... passed\n";

	//
	//	select nothing
	//
	DataFilter filter2("box(128.5,128.5,0,0,0,RAWX,RAWY)");
	assert(filter2.doSelect(*table, qualColName, 0) == 0);

	cdsel = table->column(qualColName)->data();
	sel = cdsel->int8Data();
	for (unsigned n=0; n<maxx * maxy; ++n)
		assert(sel[n] == 0);
	delete cdsel;
	std::cout<<"Filter2 test ... passed\n";

	//
	//	all events with X >= 128 - should select half of all events
	//
	DataFilter filter3("RAWX >= 128");
	assert(filter3.doSelect(*table, qualColName, 0) == maxx*maxy/2 + maxx);
	std::cout<<"Filter3 test ... passed\n";

	//
	//	select circle
	//
	DataFilter filter4("circle(1,1,128.,RAWX,RAWY)");
	assert(filter4.doSelect(*table, qualColName, 0) == 12987);
	ColumnData *cdd = table->column("DIST")->data();
	real64 *d = cdd->real64Data();
	cdsel = table->column(qualColName)->data();
	sel = cdsel->int8Data();
	for (unsigned n=0; n<maxx*maxy; ++n)
	if (!sel[n])
		assert(d[n]>128.);
	else
		assert(d[n]<=128.);
	delete cdsel;
	std::cout<<"Filter4 test ... passed\n";

	//
	//	select a ring and check number of selected events
	//
	DataFilter filter5("ring(128,128,20,100,RAWX,RAWY)");
	assert(filter5.doSelect(*table, qualColName, 0) == 30172);
	std::cout<<"Filter5 test ... passed\n";

	//
	//	a more complex expression
	//
	DataFilter filter6("CIRCLE(70.0,190.0,35.0,RAWX,RAWY)||"
					   "CIRCLE(177.0,191.5,42.0,RAWX,RAWY)||"
					   "ELLIPSE(121.5,107.0,27.0,108.0,0.0,RAWX,RAWY)");
	assert(filter6.doSelect(*table, qualColName, 0) == 18408);
	std::cout<<"Filter6 test ... passed\n";

	//
	//	another complex expression
	//
	DataFilter filter7("((100*sin(RAWX/256*360) > RAWY-200) && "
					   "(100*sin(RAWX/256*360) < RAWY-150)) || "
					   "((100*sin(RAWY/256*360) > RAWX-200) && "
					   "(100*sin(RAWY/256*360) < RAWX-150))");
	assert(filter7.doSelect(*table, qualColName, 0) == 19361);
	std::cout<<"Filter7 test ... passed\n";

	#if 0
	std::vector<RowRange> rowRangeVec;
	filter7.doSelect(*table, rowRangeVec);
	std::cout<<"number of selected rows: "<<rowRangeVec.size()<<std::endl;
	for (unsigned n=0; n<rowRangeVec.size(); ++n)
		std::cout<<n<<" "<<rowRangeVec[n].startRow<<" - "<<rowRangeVec[n].stopRow
			<<std::endl;
	#endif

	//
	//	and another one ...
	//
	DataFilter filter8("(CIRCLE(86.0,192.5,15.0,RAWX,RAWY)||"
					   "CIRCLE(152.5,194.0,14.5,RAWX,RAWY)||ELLIPSE"
					   "(119.0,151.5,4.5,21.5,0.0,RAWX,RAWY)||"
					   "BOX(102.0,106.0,7.0,14.0,0.0,RAWX,RAWY)||"
					   "BOX(126.0,104.0,7.0,13.0,0.0,RAWX,RAWY)||"
					   "ring(120,160,65,75,RAWX,RAWY))");
	assert(filter8.doSelect(*table, qualColName, 0) == 6761);
	std::cout<<"Filter8 test ... passed\n";

	//
	//	do filtering of photons which survived filter8
	//
	DataFilter filter9("SELECTED");
	assert(filter9.doSelect(*table, "X", -1) == 6761);
	std::cout<<"Filter9 test ... passed\n";

	DataFilter filter10("!SELECTED");
	assert(filter10.doSelect(*table, "Y", 14) == maxx*maxy-6761);
	std::cout<<"Filter10 test ... passed\n";

	DataFilter filter11("SELECTED");
	assert(filter11.doSelect(*table, qualColName, 0) == maxx*maxy-6761);
	std::cout<<"Filter11 test ... passed\n";

	DataFilter filter12("SELECTED || !SELECTED");
	assert(filter12.doSelect(*table, qualColName, 0) == maxx*maxy);
	std::cout<<"Filter12 test ... passed\n";

	table->deleteColumn("X");
	table->deleteColumn("Y");

	//
	//	repeat test #4 this time using attributes
	//
	std::string
		expr(DataFilter::expandSelectionExpression("#DIST <= 128.",*table));
	DataFilter filter13(expr);
	assert(filter13.doSelect(*table, qualColName, 0) == 12987);
	std::cout<<"Filter13 test ... passed\n";

	std::string
		expr2(DataFilter::expandSelectionExpression("#DISTSEL", *table));
	DataFilter filter14(expr);
	assert(filter14.doSelect(*table, qualColName, 0) == 12987);
	std::cout<<"Filter14 test ... passed\n";

	//	we do not need to test the doIt() method which returns the
	//	row ranges because the other doIt() makes use of this, thus,
	//	if the latter is ok, so should be the former


	//	tidy up
	delete cdx;
	delete cdy;
	delete cdb;
	delete cdd;
	dataSetServer->close(set);

	//
	//	a final test with destructive filtering
	//
    system("cp filtertest.ds f1.ds");
	set = dataSetServer->open("f1.ds", DataSet::Modify);
	table = set->table("SPATIAL");
	DataFilter filter15("ring(128,128,20,100,RAWX,RAWY)");
	assert(filter15.doFilter(*table) == 30172);
	std::cout<<"Filter15 test ... passed\n";
	dataSetServer->close(set);
    unlink("f1.ds");

	std::cout<<"All tests successfully passed.\n";
	return 0;
}

